/* cquant.c */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include "loadimg.h"

/* partially based on colorquant.c, see copyright below
   significantly modifed: 
   - changed C style to arahne's
   - added function prototypes
   - removed unused parameters
   - changed the format of input data (from separate r, g, b images to one rgb image, added support for line padding)
   - changed the format of resulting colormap (from separated r, g, b, colormap arrays to one rgb array)
   - optimized the functions to fixed precision of quantization tables (5 significant bits per R, G, B)
   - increased precison of "long" to "long long" to correctly handle images 5000x5000 and higher
     especially on low number of colors, where precision was overflowing on the original version
   - increased precision of colormap generation to 6 bits per r, g, b
   modification done on 12-Mar-2007 by Dusan Peterc, arahne@arahne.si
*/

/*
 * This software is copyrighted as noted below.  It may be freely copied,
 * modified, and redistributed, provided that the copyright notice is 
 * preserved on all copies.
 * 
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is".  Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a program or other software product
 * without supplying the source, or without informing the end-user that the 
 * source is available for no extra charge.
 *
 * If you modify this software, you should include a notice giving the
 * name of the person performing the modification, the date of modification,
 * and the reason for such modification.

 * Perform variance-based color quantization on a "full color" image.
 * Author:	Craig Kolb
 *		Department of Mathematics
 *		Yale University
 *		kolb@yale.edu
 * Date:	Tue Aug 22 1989
 * Copyright (C) 1989 Craig E. Kolb
*/

#define NORMALIZE_BITS	2 /* shift byte left by this many bits to get our normal size */ 
#define SIGNIF_BITS	6 /* number of significant bits in our quantization */
#define NORMALIZE_AND 252 /* and with this to strip off insignificant bits */

#include <math.h>

/* Flag bits.
 *
 * CQ_FAST:	If set, the rgbmap will be constructed quickly.  If
 *		not set, the rgbmap will be built much slower, but
 *		more accurately.  In most cases, CQ_FAST should be
 *		set, as the error introduced by the approximation is
 *		usually small.
 * CQ_QUANTIZE:	If set, the data in red, green, and blue is not
 * 		pre-quantized, and will be quantized "on the fly".
 * 		This slows the routine slightly.  If not set, the data
 * 		has been prequantized to 'bits' significant bits.*/
#define CQ_FAST		1
#define CQ_QUANTIZE	2
#define CQ_NO_RGBMAP	4


/* Ansi uses HUGE_VAL. */
#ifndef HUGE
#ifdef HUGE_VAL
#define HUGE HUGE_VAL
#else
#define HUGE MAXFLOAT
#endif
#endif

/* Maximum number of colormap entries.  To make larger than 2^8, the rgbmap
 * type will have to be changed from unsigned chars to something larger. */
#define MAXCOLORS		256
/* Value corrersponding to full intensity in colormap.  The values placed
 * in the colormap are scaled to be between zero and this number.  Note
 * that anything larger than 255 is going to lead to problems, as the
 * colormap is declared as an unsigned char. */
#define FULLINTENSITY		255
#define MAX(x,y)	((x) > (y) ? (x) : (y))

/* Readability constants. */
#define REDI		0	
#define GREENI		1
#define BLUEI		2	
#define TRUE		1
#define FALSE		0

typedef struct
 {
  double	weightedvar;		/* weighted variance */
  float		mean[3];		/* centroid */
  unsigned long long weight,		/* # of pixels in box */
		freq[3][MAXCOLORS];	/* Projected frequencies */
  int 		low[3], high[3];	/* Box extent */
 } Box;

static unsigned long long *Histogram,	/* image histogram */
			SumPixels;	/* total # of pixels */
static unsigned int	ColormaxI;	/* # of colors, 2^Bits */
static Box		*Boxes;		/* Array of color boxes. */

static void QuantHistogramTrue(unsigned char *trueData, int trueBPL, int sizeX, int sizeY, Box *box);
static void QuantHistogramCmap(unsigned char *cmapData, int cmapBPL, int sizeX, int sizeY, unsigned char *cmap, Box *box);
static int CutBoxes(Box	*boxes, int colors);
static void ComputeRGBMap(Box *boxes, int colors, unsigned char *rgbmap, int fast);
static void BoxStats(Box *box);
static int CutBox(Box *box, Box *newbox);
static int GreatestVariance(Box *boxes, int n);
static int FindCutpoint(Box *box, int color, Box *newbox1, Box *newbox2);
static void UpdateFrequencies(Box *box1, Box *box2);
static void SetRGBmap(int boxnum, Box *box, unsigned char *rgbmap);
static void find_colors(Box *boxes, int colors, unsigned char *rgbmap, int otherimages);
static int getneighbors(Box *boxes, int num, int *neighbors, int colors);
static void makenearest(Box *boxes, int boxnum, int nneighbors, int *neighbors,
	unsigned char *rgbmap, int otherimages);

/* Perform variance-based color quantization on a 24-bit image.
 *
 *	rgbmap			An array of unsigned chars of size (2^bits)^3.
 *				This array is used to map from pixels to
 *				colormap entries.  The 'prequantized' red,
 *				green and blue components of a pixel
 *				are used as an index into rgbmap to retrieve
 *				the index which should be used into the colormap
 *				to represent the pixel.  In short:
 *				index = rgbmap[(((r << bits) | g) << bits) | b];
 * 	flags			A set of bit-flags:
 * 		CQ_FAST		If set, the rgbmap will be constructed
 *				quickly.  If not set, the rgbmap will
 *				be built more slowly, but more
 *				accurately.  In most cases, fast
 *				should be set, as the error introduced
 *				by the approximation is usually small.
 * 		CQ_QUANTIZE:	If set, the data in red, green, and
 * 				blue is not pre-quantized, and will be
 * 				quantized "on the fly".  This slows
 * 				the routine slightly.  If not set, the
 * 				data has been prequantized to 'bits'
 * 				significant bits. */

static int InitHistogram(int colors)
{
 ColormaxI = 1 << SIGNIF_BITS;	/* 2 ^ significant bits */

 Histogram = (unsigned long long *)calloc(ColormaxI*ColormaxI*ColormaxI, sizeof(long long));
 Boxes = (Box *)malloc(colors * sizeof(Box));
 if (!Histogram || !Boxes)
  return(FALSE);
 /* Zero-out the projected frequency arrays of the largest box. */
 memset(Boxes->freq[0], 0, ColormaxI * sizeof(unsigned long long));
 memset(Boxes->freq[1], 0, ColormaxI * sizeof(unsigned long long));
 memset(Boxes->freq[2], 0, ColormaxI * sizeof(unsigned long long));
 return(TRUE);
}

static void FreeHistogram()
{
 free((char *)Histogram);
 free((char *)Boxes);
}

/* Compute the histogram of the true color image as well as the projected frequency
 * arrays for the first world-encompassing box. */
static void QuantHistogramTrue(unsigned char *trueData, int trueBPL, int sizeX, int sizeY, Box *box)
{
 register unsigned long long *rf, *gf, *bf;
 int i, j;
 unsigned char rr, gg, bb;
 unsigned char *tData;

 rf = box->freq[0];
 gf = box->freq[1];
 bf = box->freq[2];

 for (i = 0; i < sizeY; i++)
  {
   tData = &trueData[i*trueBPL];
   for (j = 0; j < sizeX; j++)
    {
     rr = (unsigned char) ((*tData++)>>NORMALIZE_BITS);
     gg = (unsigned char) ((*tData++)>>NORMALIZE_BITS);
     bb = (unsigned char) ((*tData++)>>NORMALIZE_BITS);
     rf[rr]++;
     gf[gg]++;
     bf[bb]++;
     Histogram[(((rr<<SIGNIF_BITS)|gg)<<SIGNIF_BITS)|bb]++;
    }
  }
}

/* Compute the histogram of the colormap image as well as the projected frequency
 * arrays for the first world-encompassing box. */
static void QuantHistogramCmap(unsigned char *cmapData, int cmapBPL, int sizeX, int sizeY, unsigned char *cmap, Box *box)
{
 register unsigned long long *rf, *gf, *bf;
 int i, j;
 unsigned char *cData, *cmapPtr;
 unsigned char cmap2[3*256];
 unsigned long long chist[256];

 /* make a copy of colormap and normalize it so we don't have to do it in the main loop */
 for (i=0; i<3*256; i++)
  cmap2[i] = (unsigned char) (cmap[i]>>NORMALIZE_BITS);
 memset(chist, 0, sizeof(unsigned long long)*256);

 rf = box->freq[0];
 gf = box->freq[1];
 bf = box->freq[2];

 /* make a colormap histogram */
 for (i = 0; i < sizeY; i++)
  {
   cData = &cmapData[i*cmapBPL];
   for (j = sizeX; j; j--)
    chist[*cData++]++;
  }

 /* then fill in the color reduction histogram based on the colormap histogram */
 for (i = 0; i < 256; i++)
  if (chist[i])
   {
    cmapPtr = &cmap2[i*3];
    rf[cmapPtr[0]]+=chist[i];
    gf[cmapPtr[1]]+=chist[i];
    bf[cmapPtr[2]]+=chist[i];
    Histogram[(((cmapPtr[0]<<SIGNIF_BITS)|cmapPtr[1])<<SIGNIF_BITS)|cmapPtr[2]]+=chist[i];
   }
}

/* Interatively cut the boxes. */
static int CutBoxes(Box	*boxes, int colors)
{
 int curbox;

 boxes[0].low[REDI] = boxes[0].low[GREENI] = boxes[0].low[BLUEI] = 0;
 boxes[0].high[REDI] = boxes[0].high[GREENI] = boxes[0].high[BLUEI] = ColormaxI;
 boxes[0].weight = SumPixels;
 BoxStats(&boxes[0]);

 for (curbox = 1; curbox < colors; curbox++)
  if (CutBox(&boxes[GreatestVariance(boxes, curbox)],  &boxes[curbox]) == FALSE)
   break;
 return curbox;
}

/* Return the number of the box in 'boxes' with the greatest variance.
 * Restrict the search to those boxes with indices between 0 and n-1. */
static int GreatestVariance(Box *boxes, int n)
{
 register int i, whichbox = 0;
 double max;

 max = -1;
 for (i = 0; i < n; i++)
  if (boxes[i].weightedvar > max)
   {
    max = boxes[i].weightedvar;
    whichbox = i;
   }
 return(whichbox);
}

/* Compute mean and weighted variance of the given box. */
static void BoxStats(Box *box)
{
 register int i, color;
 unsigned long long *freq;
 double mean, var;

 if(box->weight == 0)
  {
   box->weightedvar = 0;
   return;
  }

 box->weightedvar = 0.;
 for (color = 0; color < 3; color++)
  {
   var = mean = 0;
   i = box->low[color];
   freq = &box->freq[color][i];
   for (; i < box->high[color]; i++, freq++)
    {
     mean += i * *freq;
     var += i*i* *freq;
    }
   box->mean[color] = (float) (mean / (double)box->weight);
   box->weightedvar += var - box->mean[color]*box->mean[color]*(double)box->weight;
  }
 box->weightedvar /= SumPixels;
}

/* Cut the given box.  Returns TRUE if the box could be cut, FALSE otherwise. */
static int CutBox(Box *box, Box *newbox)
{
 int i;
 double totalvar[3];
 Box newboxes[3][2];

 /* Can't cut this box. */
 if ((box->weightedvar < 0.00001) || (box->weight == 0))
  return(FALSE);

 /* Find 'optimal' cutpoint along each of the red, green and blue
  * axes.  Sum the variances of the two boxes which would result
  * by making each cut and store the resultant boxes for 
  * (possible) later use. */
 for (i = 0; i < 3; i++)
  if (FindCutpoint(box, i, &newboxes[i][0], &newboxes[i][1]))
   totalvar[i] = newboxes[i][0].weightedvar + newboxes[i][1].weightedvar;
  else
   totalvar[i] = HUGE;

 /* Find which of the three cuts minimized the total variance
  * and make that the 'real' cut. */
 if (totalvar[REDI] <= totalvar[GREENI] && totalvar[REDI] <= totalvar[BLUEI])
  {
   *box = newboxes[REDI][0];
   *newbox = newboxes[REDI][1];
  }
 else
  if (totalvar[GREENI] <= totalvar[REDI] && totalvar[GREENI] <= totalvar[BLUEI])
   {
    *box = newboxes[GREENI][0];
    *newbox = newboxes[GREENI][1];
   }
  else
   {
    *box = newboxes[BLUEI][0];
    *newbox = newboxes[BLUEI][1];
   }
 return(TRUE);
}

/* Compute the 'optimal' cutpoint for the given box along the axis
 * indcated by 'color'.  Store the boxes which result from the cut
 * in newbox1 and newbox2. */
static int FindCutpoint(Box *box, int color, Box *newbox1, Box *newbox2)
{
 float u, v, max;
 int i, maxindex, minindex, cutpoint;
 unsigned long long optweight, curweight;

 if (box->low[color] + 1 == box->high[color])
	 return FALSE;	/* Cannot be cut. */
 minindex = (int)((box->low[color] + box->mean[color]) * 0.5);
 maxindex = (int)((box->mean[color] + box->high[color]) * 0.5);

 cutpoint = minindex;
 optweight = box->weight;

 curweight = 0;
 for (i = box->low[color] ; i < minindex ; i++)
	 curweight += box->freq[color][i];
 u = 0.;
 max = -1;
 for (i = minindex; i <= maxindex ; i++)
  {
   curweight += box->freq[color][i];
   if (curweight == box->weight)
    break;
   u += (float)(i * box->freq[color][i]) / (float)box->weight;
   v = ((float)curweight / (float)(box->weight-curweight)) * (box->mean[color]-u)*(box->mean[color]-u);
   if (v > max)
    {
     max = v;
     cutpoint = i;
     optweight = curweight;
    }
  }
 cutpoint++;
 *newbox1 = *newbox2 = *box;
 newbox1->weight = optweight;
 newbox2->weight -= optweight;
 newbox1->high[color] = cutpoint;
 newbox2->low[color] = cutpoint;
 UpdateFrequencies(newbox1, newbox2);
 BoxStats(newbox1);
 BoxStats(newbox2);

 return(TRUE);	/* Found cutpoint. */
}

/* Update projected frequency arrays for two boxes which used to be  a single box. */
static void UpdateFrequencies(Box *box1, Box *box2)
{
 register unsigned long long myfreq, *h;
 register int b, g, r;
 int roff;

 memset(box1->freq[0], 0, ColormaxI * sizeof(unsigned long long));
 memset(box1->freq[1], 0, ColormaxI * sizeof(unsigned long long));
 memset(box1->freq[2], 0, ColormaxI * sizeof(unsigned long long)); 

 for (r = box1->low[0]; r < box1->high[0]; r++)
  {
   roff = r << SIGNIF_BITS;
   for (g = box1->low[1];g < box1->high[1]; g++)
    {
     b = box1->low[2];
     h = Histogram + (((roff | g) << SIGNIF_BITS) | b);
     for (; b < box1->high[2]; b++)
      {
       if ((myfreq = *h++) == 0)
	continue;
       box1->freq[0][r] += myfreq;
       box1->freq[1][g] += myfreq;
       box1->freq[2][b] += myfreq;
       box2->freq[0][r] -= myfreq;
       box2->freq[1][g] -= myfreq;
       box2->freq[2][b] -= myfreq;
      }
    }
  }
}

/* Compute RGB to colormap index map. */
static void ComputeRGBMap(Box *boxes, int colors, unsigned char *rgbmap, int fast)
{
 register int i;

 if (fast)
  /* The centroid of each box serves as the representative
   * for each color in the box. */
  for (i = 0; i < colors; i++)
   SetRGBmap(i, &boxes[i], rgbmap);
 else
  /* Find the 'nearest' representative for each pixel. */
  find_colors(boxes, colors, rgbmap, 1);
}

/* Make the centroid of "boxnum" serve as the representative for
 * each color in the box. */
static void SetRGBmap(int boxnum, Box *box, unsigned char *rgbmap)
{
 register int r, g, b;

 for (r = box->low[REDI]; r < box->high[REDI]; r++)
  for (g = box->low[GREENI]; g < box->high[GREENI]; g++)
   for (b = box->low[BLUEI]; b < box->high[BLUEI]; b++)
    rgbmap[(((r<<SIGNIF_BITS)|g)<<SIGNIF_BITS)|b]=(char)boxnum;
}

/* Form colormap and NearestColor arrays. */
static void find_colors(Box *boxes, int colors, unsigned char *rgbmap, int otherimages)
{
 register int i;
 int num, *neighbors;

 neighbors = (int *)malloc(colors * sizeof(int));

 /* Form map of representative (nearest) colors. */
 for (i = 0; i < colors; i++)
  {
   /* Create list of candidate neighbors and
    * find closest representative for each
    * color in the box. */
   num = getneighbors(boxes, i, neighbors, colors);
   makenearest(boxes, i, num, neighbors, rgbmap, otherimages);
  }
 free((char *)neighbors);
}

/* In order to minimize our search for 'best representative', we form the
 * 'neighbors' array.  This array contains the number of the boxes whose
 * centroids *might* be used as a representative for some color in the
 * current box.  We need only consider those boxes whose centroids are closer
 * to one or more of the current box's corners than is the centroid of the
 * current box. 'Closeness' is measured by Euclidean distance. */
static int getneighbors(Box *boxes, int num, int *neighbors, int colors)
{
 register int i, j;
 Box *bp;
 float dist, LowR, LowG, LowB, HighR, HighG, HighB, ldiff, hdiff;

 bp = &boxes[num];

 ldiff = bp->low[REDI] - bp->mean[REDI];
 ldiff *= ldiff;
 hdiff = bp->high[REDI] - bp->mean[REDI];
 hdiff *= hdiff;
 dist = MAX(ldiff, hdiff);

 ldiff = bp->low[GREENI] - bp->mean[GREENI];
 ldiff *= ldiff;
 hdiff = bp->high[GREENI] - bp->mean[GREENI];
 hdiff *= hdiff;
 dist += MAX(ldiff, hdiff);

 ldiff = bp->low[BLUEI] - bp->mean[BLUEI];
 ldiff *= ldiff;
 hdiff = bp->high[BLUEI] - bp->mean[BLUEI];
 hdiff *= hdiff;
 dist += MAX(ldiff, hdiff);

 dist = (float)sqrt((double)dist);

 /* Loop over all colors in the colormap, the ith entry of which
  * corresponds to the ith box.
  * If the centroid of a box is as close to any corner of the
  * current box as is the centroid of the current box, add that
  * box to the list of "neighbors" of the current box. */
 HighR = (float)bp->high[REDI] + dist;
 HighG = (float)bp->high[GREENI] + dist;
 HighB = (float)bp->high[BLUEI] + dist;
 LowR = (float)bp->low[REDI] - dist;
 LowG = (float)bp->low[GREENI] - dist;
 LowB = (float)bp->low[BLUEI] - dist;
 for (i = j = 0, bp = boxes; i < colors; i++, bp++)
  if (LowR <= bp->mean[REDI] && HighR >= bp->mean[REDI] &&
      LowG <= bp->mean[GREENI] && HighG >= bp->mean[GREENI] &&
      LowB <= bp->mean[BLUEI] && HighB >= bp->mean[BLUEI])
	  neighbors[j++] = i;

 return(j); /* Return the number of neighbors found. */
}

/* Assign representative colors to every pixel in a given box through
 * the construction of the NearestColor array.  For each color in the
 * given box, we look at the list of neighbors passed to find the
 * one whose centroid is closest to the current color. */
static void makenearest(Box *boxes, int boxnum, int nneighbors, int *neighbors,
	unsigned char *rgbmap, int otherimages)
{
 register int n, b, g, r;
 double rdist, gdist, bdist, dist, mindist;
 int which = 0, *np;
 Box *box;

 box = &boxes[boxnum];

 for (r = box->low[REDI]; r < box->high[REDI]; r++)
  {
   for (g = box->low[GREENI]; g < box->high[GREENI]; g++)
    {
     for (b = box->low[BLUEI]; b < box->high[BLUEI]; b++)
      {
/* The following "if" is to avoid doing extra work when the RGBmap is
* not going to be used for other images. */
       if ((!otherimages) && (Histogram[(((r<<SIGNIF_BITS)|g)<<SIGNIF_BITS)|b] == 0))
	continue;

       mindist = HUGE;
       /* Find the colormap entry which is closest to the current color. */
       np = neighbors;
       for (n = 0; n < nneighbors; n++, np++)
	{
	 rdist = r-boxes[*np].mean[REDI];
	 gdist = g-boxes[*np].mean[GREENI];
	 bdist = b-boxes[*np].mean[BLUEI];
	 dist = rdist*rdist + gdist*gdist + bdist*bdist;
	 if (dist < mindist)
	  {
	   mindist = dist;
	   which = *np; 
	  }
	}
       /* The colormap entry closest to this color is used as a representative. */
       rgbmap[(((r<<SIGNIF_BITS)|g)<<SIGNIF_BITS)|b] = (unsigned char) which;
      }
    }
  }
}

static int findClosestPaletteRGB(unsigned char *rgb, unsigned char *cmap, int cmapColors)
{
 int dist, mindist = INT_MAX;
 int i, minpos = 0;
 unsigned char *cmapPtr = cmap;

 for (i=0; i<cmapColors; i++)
  {
   dist  = ((int)rgb[0]-*cmapPtr)*((int)rgb[0]-*cmapPtr);
   cmapPtr++;
   dist += ((int)rgb[1]-*cmapPtr)*((int)rgb[1]-*cmapPtr);
   cmapPtr++;
   dist += ((int)rgb[2]-*cmapPtr)*((int)rgb[2]-*cmapPtr);
   cmapPtr++;
   
   if (dist<mindist)
    { mindist = dist; minpos = i; }
  }
 return(minpos);
}

/* check if true color image actualy has less than 256 colors, and return color palette and number of colors */
/* or 0 (false) if it has more colors and real color reduction is nesessary */
/*
int checkTrueColorPalette(unsigned char *trueData, int trueBPL, int sizeX, int sizeY, unsigned char *cmap)
{
 unsigned char *tdata;
 int match=0, foundColors = 1;
 int indx, x, y;
 int rgbCmap[257], rgb;

 // fill in the first color, so we can easily check if equal as previous
 // on big images, it is very likely that large % of consecutive colors will be equal
 rgbCmap[0] = ((int)trueData[0]<<16) | ((int)trueData[1]<<8) | (int)trueData[2];      

 for (y=0; y<sizeY; y++)
  {
   tdata = &trueData[y*trueBPL];
   if (foundColors<=256)
    for (x=0; x<sizeX; x++)
     {
      rgb = ((int)tdata[0]<<16) | ((int)tdata[1]<<8) | (int)tdata[2];
      // check if equal as previous
      if (rgb!=rgbCmap[match])
       {
	// find exact match from colors we had so far
	for (indx=0; indx<foundColors; indx++)
	 if (rgb==rgbCmap[indx])
	  { match = indx; indx = 300; } // found and bailout

        if (indx<300) // if not found
         {
          if (foundColors<=256)
	   {
	    match = foundColors;
            rgbCmap[foundColors++]=rgb;
	   }
	  else
	   x = sizeX; // bailout
	 }
       }
      tdata+=3;
     }
   else
    y = sizeY; // bailout
  }
 if (foundColors<=256)
  {
   for (indx=0; indx<foundColors; indx++)
    {
     cmap[indx*3+0] = rgbCmap[indx]>>16;
     cmap[indx*3+1] = (rgbCmap[indx]>>8)&255;
     cmap[indx*3+2] = rgbCmap[indx]&255;
    }
   return(foundColors);
  }
 return(0);
}
*/








/* tuneable color quantization function: trueData is input data in RGB, cmapData is colormapped
   output image. You set the requested number of colors, and the program returns the actual
   number. You can previously put in the colormap the colors you want to have
   in the image (regardless of histogram), and indicate this by setting the number of fixColors>0.
   If trueData is NULL, then we are doing quantization of colormapped image.
   If you want the get the colormap without remapping, pass trueData, but no cmapData */
int colorQuantize(unsigned char *trueData, int trueBPL, unsigned char *cmapData, int cmapBPL,
 int sizeX, int sizeY, unsigned char *cmap, int fixColors, int *requestColors)
{
 long int i, j;
 unsigned char *cdata, *tdata;
 unsigned char lcmap[3*256];
 unsigned char *rgbmap=NULL;
 int doRGBmap = trueData  && cmapData;
 int Colormax;		/* quantized full-intensity */ 
 float	Cfactor;	/* Conversion factor */

 /* count number of different colors, if less than requested, than use the list of colors as colormap and map the image to that colormap */
 /* in this way, there is no color precision loss on last bits of colormap, as we don't use subsampled color cube for quantization */
 /* the color histogram and color reduction/remapping is also done in single step */
 if (!fixColors && (*requestColors<=256) && trueData && cmapData)
  {
   int match=0, foundColors = 1;
   int indx, x, y;
   int rgbCmap[257], rgb;

   /* fill in the first color, so we can easily check if equal as previous */
   /* on big images, it is very likely that large % of consecutive colors will be equal */
   rgbCmap[0] = ((int)trueData[0]<<16) | ((int)trueData[1]<<8) | (int)trueData[2];      

   for (y=0; y<sizeY; y++)
    {
     tdata = &trueData[y*trueBPL];
     cdata = &cmapData[y*cmapBPL];
     if (foundColors<=*requestColors)
      for (x=0; x<sizeX; x++)
       {
	rgb = ((int)tdata[0]<<16) | ((int)tdata[1]<<8) | (int)tdata[2];
	/* check if equal as previous */
        if (rgb!=rgbCmap[match])
	 {
	  /* find exact match from colors we had so far */
	  for (indx=0; indx<foundColors; indx++)
	   if (rgb==rgbCmap[indx])
	    { match = indx; indx = 300; } /* found and bailout */
 
          if (indx<300) /* if not found */
           {
            if (foundColors<=*requestColors)
	     {
	      match = foundColors;
              rgbCmap[foundColors++]=rgb;
	     }
	    else
	     x = sizeX; /* bailout */
	   }
	 }
        *cdata++ = match;
	tdata+=3;
       }
     else
      y = sizeY; /* bailout */
    }
//printf("reuqest %d found %d\n", *requestColors, foundColors);
   /* actual number of colors is smaller than requested; write colormap and exit */
   if (foundColors<=*requestColors)
    {
     for (i=0; i<foundColors; i++)
      {
       cmap[i*3+0] = rgbCmap[i]>>16;
       cmap[i*3+1] = (rgbCmap[i]>>8)&255;
       cmap[i*3+2] = rgbCmap[i]&255;
      }
     *requestColors = foundColors;
     return(LOAD_OK);
    }
  }

 /* make a copy of the colormap */
 memcpy(lcmap, cmap, 3*256);
 if (doRGBmap)
  rgbmap = malloc(1<<(SIGNIF_BITS*3));

 InitHistogram(*requestColors);
 Colormax = ColormaxI - 1;
 Cfactor = (float) FULLINTENSITY / Colormax;

 SumPixels = sizeX*sizeY;

 if (trueData)
  QuantHistogramTrue(trueData, trueBPL, sizeX, sizeY, &Boxes[0]);
 else if (!trueData && cmapData)
  QuantHistogramCmap(cmapData, cmapBPL, sizeX, sizeY, cmap, &Boxes[0]);

 /* "tune" the histogram so output colormap will contain the passed colors */
 if (fixColors)
  {
   register unsigned long long *rf, *gf, *bf;
   unsigned char *cmapPtr;
   unsigned char cmap2[3*fixColors];
   unsigned long long halfImage = sizeX*sizeY/2; // fix it as if half of pixels would have this color

   /* make a copy of colormap and normalize it so we don't have to do it in the main loop */
   for (i=0; i<3*fixColors; i++)
    cmap2[i] = (unsigned char) (cmap[i]>>NORMALIZE_BITS);

   rf = Boxes[0].freq[0];
   gf = Boxes[0].freq[1];
   bf = Boxes[0].freq[2];

   /* then add half of the image area to the histogram for each color */
   for (i = 0; i < fixColors; i++)
    {
     cmapPtr = &cmap2[i*3];
     rf[cmapPtr[0]]+=halfImage;
     gf[cmapPtr[1]]+=halfImage;
     bf[cmapPtr[2]]+=halfImage;
     Histogram[(((cmapPtr[0]<<SIGNIF_BITS)|cmapPtr[1])<<SIGNIF_BITS)|cmapPtr[2]] += halfImage;
     SumPixels += halfImage;
    }
  }

 /* color reduction here */
 i = CutBoxes(Boxes, *requestColors);
 *requestColors = i;

 /* We now know the set of representative colors.  We now
  * must fill in the colormap and convert the representatives
  * from their 'prequantized' range to 0-FULLINTENSITY. */
 for (i = 0; i < *requestColors; i++)
  {
   cmap[i*3    ] = (unsigned char)(Boxes[i].mean[REDI]   * Cfactor + 0.5);
   cmap[i*3 + 1] = (unsigned char)(Boxes[i].mean[GREENI] * Cfactor + 0.5);
   cmap[i*3 + 2] = (unsigned char)(Boxes[i].mean[BLUEI]  * Cfactor + 0.5);
  }

 if (doRGBmap)
  ComputeRGBMap(Boxes, *requestColors, rgbmap, CQ_QUANTIZE | CQ_FAST);

 FreeHistogram();

 /* now remap image data, if we have them */

 if (cmapData && trueData) /* do the mapping from true color to cmap image */
  {
   int index;
   unsigned char *rgbmapBits;

   /* apparnetly rgbmap does not contain 100% correct mapping */
   /* so we need to calculate the mapping of pixel to reduced color again */
   /* but to avoid scanning all colors for each pixel, we use rgbmapBits to remember which points did we already calculate */
   /* overall, the performance penalty is not big, and now we are sure to get correct mapping to the palette */
   /* the incorrect rgbmap could also indicate errors in color reduction algorithm, but it is too hard to debug ;-( */

   rgbmapBits = calloc(((1<<(SIGNIF_BITS*3))+7)>>3, sizeof(unsigned char));
   if (!rgbmapBits)
    return(FALSE);

   for(i=0; i<sizeY; i++)
    {
     tdata = &trueData[i*trueBPL];
     cdata = &cmapData[i*cmapBPL];
     for(j=0; j<sizeX; j++, tdata+=3)
//      *cdata++ =  (unsigned char) findClosestPaletteRGB(tdata, cmap, *requestColors); /* yikes! very slow */
//      *cdata++ = rgbmap[((((tdata[0]>>NORMALIZE_BITS)<<SIGNIF_BITS) | (tdata[1]>>NORMALIZE_BITS))<<SIGNIF_BITS) | (tdata[2]>>NORMALIZE_BITS)]; // it looks that rgbmap is incorrect after color recution
      {
       index = ((((tdata[0]>>NORMALIZE_BITS)<<SIGNIF_BITS) | (tdata[1]>>NORMALIZE_BITS))<<SIGNIF_BITS) | (tdata[2]>>NORMALIZE_BITS);
       if (!(rgbmapBits[index>>3]&(1<<(index&7))))
        {
// looks like we should not normalize tdata, only index needs to be normalized
//	 tdata[0]&=NORMALIZE_AND;
//	 tdata[1]&=NORMALIZE_AND;
//	 tdata[2]&=NORMALIZE_AND;
	 rgbmap[index] = (unsigned char) findClosestPaletteRGB(tdata, cmap, *requestColors);
         rgbmapBits[index>>3] |= (1<<(index&7));
	}
       *cdata++ = rgbmap[index];
      }
    }
   free(rgbmap);
   free(rgbmapBits);
  }
 if (cmapData && !trueData)
  {
   unsigned char transTable[256];
   unsigned long limit = sizeY*cmapBPL;

   /* map the original colormap to resulting colormap */
   for (i=0; i<256; i++)
    transTable[i] = (unsigned char) findClosestPaletteRGB(&lcmap[i*3], cmap, *requestColors);
   /* now just remap the data */
   cdata = cmapData;
   while (limit--)
    { *cdata=transTable[*cdata]; cdata++; }
  }

 return(LOAD_OK);
}


#define RAND_AND	32767 /* must be (2 to the power of n) - 1 */

/* remap a true color image to colormap image, using the passed colormap, without any colormap compression */
/* dither noise controls the level of random dithering, valid values ar 0-100 */
int trueColorRemapToCmap(unsigned char *trueData, int trueBPL, unsigned char *cmapData, int cmapBPL,
 int sizeX, int sizeY, unsigned char *cmap, int cmapColors, int ditherNoise)
{
 int x, y, i;
 unsigned char *srcData, *dstData;
 char randTable[RAND_AND+1];
 unsigned int randPos=0;
 unsigned char rgb[3];
 unsigned char *rgbmap, *rgbmapBits;
 int index;
 unsigned char truncateArray[2048], *truncateA;

 if ((!trueData) || (!cmapData) || (!cmap))
  return(FALSE);
 ditherNoise = min(max(0, ditherNoise), 100);

 rgbmap = malloc(1<<(SIGNIF_BITS*3));
 if (!rgbmap)
  return(FALSE);

 rgbmapBits = calloc(((1<<(SIGNIF_BITS*3))+7)/8, sizeof(unsigned char));
 if (!rgbmapBits)
  { free(rgbmap); return(FALSE); }

 if (ditherNoise)
  {
   for (i=0; i<=RAND_AND; i++)
    randTable[i] = (char) (((rand()>>3)%(ditherNoise*2+1)) - ditherNoise);

   memset(truncateArray, 0, 1024); /* set bottom half to 0 */
   memset(&truncateArray[1024], 255, 1024); /* set upper half to 255 */
   truncateA = &truncateArray[1024]; /* we will use this one for lookup */
   /* now set the usable range */
   for (x=0; x<256; x++)
    truncateA[x] = (unsigned char) x;
  } 

 for (y=0; y<sizeY; y++)
  {
   srcData = &trueData[y*trueBPL];
   dstData = &cmapData[y*cmapBPL];
   randPos = (unsigned short) rand();
   /* duplicate the main loop to avoid useless dither, if not needed */
   if (ditherNoise)
    for (x=0; x<sizeX; x++)
     {
      rgb[0] = truncateA[*srcData++ + randTable[(randPos++) & RAND_AND]];
      rgb[1] = truncateA[*srcData++ + randTable[(randPos++) & RAND_AND]];
      rgb[2] = truncateA[*srcData++ + randTable[(randPos++) & RAND_AND]];
      index = ((((rgb[0]>>NORMALIZE_BITS)<<SIGNIF_BITS) | (rgb[1]>>NORMALIZE_BITS))<<SIGNIF_BITS) | (rgb[2]>>NORMALIZE_BITS);
      if (!(rgbmapBits[index>>3]&(1<<(index&7))))
       {
	rgbmap[index] = (unsigned char) findClosestPaletteRGB(rgb, cmap, cmapColors);
	rgbmapBits[index>>3] |= (1<<(index&7));
       }
      *dstData++ = rgbmap[index];
     }
    else
     for (x=0; x<sizeX; x++)
      {
       index = ((((srcData[0]>>NORMALIZE_BITS)<<SIGNIF_BITS) | (srcData[1]>>NORMALIZE_BITS))<<SIGNIF_BITS) | (srcData[2]>>NORMALIZE_BITS);
       if (!(rgbmapBits[index>>3]&(1<<(index&7))))
	{
	 rgbmap[index] = (unsigned char) findClosestPaletteRGB(srcData, cmap, cmapColors);
	 rgbmapBits[index>>3] |= (1<<(index&7));
	}
       *dstData++ = rgbmap[index];
       srcData++; srcData++; srcData++;
      }
  }
 free(rgbmap);
 free(rgbmapBits);
 return(TRUE);
}

/* remap a true color image to palette image using given colormap and Floyd/Steinberg dithering */
/* for finding closest color, use color cube mapping instead of search through all palette colors for each pixel */
/* ditherIntensity is 16-0, 16 is normal dither, puts less dithering transition, and 0 is no dithering */
/* ditherNoise is level of noise for dithering, 0 for no noise, 100 is maximum; will add random values of +-100 */ 
void colorQuantizeDitherFS(unsigned char *trueData, int trueBPL, unsigned char *cmapData, int cmapBPL,
 int srcX, int srcY, int dstX, int dstY, unsigned char *cmap, int cmapColors, int ditherIntensity, int ditherNoise)
{
 int x,y,c;
 unsigned char *srcPtr, *dstPtr, *srcPtrRight, *srcPtrDown, *srcPtrDownRight, *srcPtrDownLeft;
 int error, index;
 unsigned int remapX[dstX+2]; /* do not make it unsigned short, as it contains values of max pixel multiplied by 3 */ 
 /* use these buffers to allow image scaling during dithering, and to distribute dithering error */
 /* make them one pixel bigger on the left and right, so we can ignore the border checking in the algorithm */
 unsigned char currLine[(dstX+2)*3], nextLine[(dstX+2)*3];
 unsigned char *rgbmap, *rgbmapBits;
 unsigned char truncateArray[2048], *truncateA;
 char randTable[RAND_AND+1];
 unsigned short randPos=0;

 ditherNoise = min(max(0, ditherNoise), 100);
 ditherIntensity = min(max(0, ditherIntensity), 16);

 memset(truncateArray, 0, 1024); /* set bottom half to 0 */
 memset(&truncateArray[1024], 255, 1024); /* set upper half to 255 */
 truncateA = &truncateArray[1024]; /* we will use this one for lookup */
 /* now set the usable range */
 for (x=0; x<256; x++)
  truncateA[x] = (unsigned char) x;

 /* need to do %, not >> and &, since we need odd number of values, for example 3 (-1, 0, 1) */
 /* anyway, it is just for the size of the table, so it is OK */
 if (ditherNoise)
  for (x=0; x<=RAND_AND; x++)
   randTable[x] = (char) (((rand()>>3)%(ditherNoise*2+1)) - ditherNoise);

 for (x=0; x<dstX; x++)
  remapX[x+1] =((x*srcX)/dstX)*3;
 /* take care of the border pixels, just duplicate the border pixels */
 remapX[0]= remapX[1];
 remapX[dstX+1]= remapX[dstX];
 rgbmap = malloc(1<<(SIGNIF_BITS*3));
 rgbmapBits = calloc(((1<<(SIGNIF_BITS*3))+7)/8, sizeof(unsigned char));

 /* initialise the first line */
 srcPtr = trueData;
 dstPtr = currLine;
 for (x=0; x<dstX+2; x++)
  {
   *dstPtr++ = srcPtr[remapX[x]];
   *dstPtr++ = srcPtr[remapX[x]+1];
   *dstPtr++ = srcPtr[remapX[x]+2];
  }

 for (y=0; y<dstY; y++)
  {
   /* initialise the next line */
   if (y + 1 < dstY)
    {
     /* feed in the next line */
     srcPtr = &trueData[trueBPL*(((y+1)*srcY)/dstY)];
     dstPtr = nextLine;
     if (srcX==dstX)
      {
       memcpy(&dstPtr[3], srcPtr, srcX*3);
       memcpy(&dstPtr[0], srcPtr, 3); /* duplicate first pixel */
       memcpy(&dstPtr[srcX*3+3], &srcPtr[(srcX-1)*3], 3); /* duplicate last pixel */
      }
     for (x=0; x<dstX+2; x++)
      {
       *dstPtr++ = srcPtr[remapX[x]];
       *dstPtr++ = srcPtr[remapX[x]+1];
       *dstPtr++ = srcPtr[remapX[x]+2];
      }
    }
   /* else we already have the next line equal to the last line, so no need to copy it again */

   dstPtr = &cmapData[cmapBPL*y];
   srcPtr = &currLine[3];
   srcPtrRight = srcPtr + 3;
   srcPtrDown = &nextLine[3];
   srcPtrDownRight = srcPtrDown + 3;
   srcPtrDownLeft =  srcPtrDown - 3;
   randPos = rand();
   for (x=dstX; x; x--)
    {
     index = ((((srcPtr[0]>>NORMALIZE_BITS)<<SIGNIF_BITS) | (srcPtr[1]>>NORMALIZE_BITS))<<SIGNIF_BITS) | (srcPtr[2]>>NORMALIZE_BITS);
     if (!(rgbmapBits[index>>3]&(1<<(index&7))))
      {
       rgbmap[index] = (unsigned char) findClosestPaletteRGB(srcPtr, cmap, cmapColors);
       rgbmapBits[index>>3] |= (1<<(index&7));
      }
     *dstPtr = rgbmap[index];
     for (c=0; c<3; ++c)
      {
       error = ((int)*srcPtr) - cmap[*dstPtr*3+c];
       if (ditherNoise)
        error += randTable[(randPos++)&RAND_AND];
       if (ditherIntensity==16)
        {
	 *srcPtrRight     = truncateA[*srcPtrRight     + ((error*7) >> 4)];
	 *srcPtrDownLeft  = truncateA[*srcPtrDownLeft  + ((error*3) >> 4)];
	 *srcPtrDown      = truncateA[*srcPtrDown      + ((error*5) >> 4)];
	 *srcPtrDownRight = truncateA[*srcPtrDownRight + ((error*1) >> 4)];
	}
       else
        {
	 *srcPtrRight     = truncateA[*srcPtrRight     + ((error*7*ditherIntensity) >> 8)];
	 *srcPtrDownLeft  = truncateA[*srcPtrDownLeft  + ((error*3*ditherIntensity) >> 8)];
	 *srcPtrDown      = truncateA[*srcPtrDown      + ((error*5*ditherIntensity) >> 8)];
	 *srcPtrDownRight = truncateA[*srcPtrDownRight + ((error*1*ditherIntensity) >> 8)];
	}
       srcPtr++;
       srcPtrRight++;
       srcPtrDown++;
       srcPtrDownRight++;
       srcPtrDownLeft++;
      }
     dstPtr++;
    }
   /* left right scanning does not fix the patterns... tried that, did not work... we rather use noise parameter */

   /* copy next line to current line, regardless of limits, so we duplicate the last line */
   memcpy(currLine, nextLine, (dstX+2)*3);
  }
 free(rgbmap);
 free(rgbmapBits);
}

/* same as above, but with dithering when we have more than 256 input colors; will  give result as 16 bit pallete index, instead of 8 bit */
/* so cmapData is unsigned short instead of unsigned char */ 
void colorQuantizeDitherFS16(unsigned char *trueData, int trueBPL, unsigned short *cmapData, int cmapBPL,
 int srcX, int srcY, int dstX, int dstY, unsigned char *cmap, int cmapColors, int ditherIntensity, int ditherNoise)
{
 int x,y,c;
 unsigned char *srcPtr, *dstPtr, *srcPtrRight, *srcPtrDown, *srcPtrDownRight, *srcPtrDownLeft;
 unsigned short *dstPtr16;
 int error, index;
 unsigned int remapX[dstX+2]; /* do not make it unsigned short, as it contains values of max pixel multiplied by 3 */ 
 /* use these buffers to allow image scaling during dithering, and to distribute dithering error */
 /* make them one pixel bigger on the left and right, so we can ignore the border checking in the algorithm */
 unsigned char currLine[(dstX+2)*3], nextLine[(dstX+2)*3];
 unsigned char *rgbmapBits;
 unsigned short *rgbmap16; /* need to have it local with 16 bpp insted of 8 bits */
 unsigned char truncateArray[2048], *truncateA;
 char randTable[RAND_AND+1];
 unsigned short randPos=0;

 ditherNoise = min(max(0, ditherNoise), 100);
 ditherIntensity = min(max(0, ditherIntensity), 16);

 memset(truncateArray, 0, 1024); /* set bottom half to 0 */
 memset(&truncateArray[1024], 255, 1024); /* set upper half to 255 */
 truncateA = &truncateArray[1024]; /* we will use this one for lookup */
 /* now set the usable range */
 for (x=0; x<256; x++)
  truncateA[x] = (unsigned char) x;

 /* need to do %, not >> and &, since we need odd number of values, for example 3 (-1, 0, 1) */
 /* anyway, it is just for the size of the table, so it is OK */
 if (ditherNoise)
  for (x=0; x<=RAND_AND; x++)
   randTable[x] = (char) (((rand()>>3)%(ditherNoise*2+1)) - ditherNoise);

 for (x=0; x<dstX; x++)
  remapX[x+1] =((x*srcX)/dstX)*3;
 /* take care of the border pixels, just duplicate the border pixels */
 remapX[0]= remapX[1];
 remapX[dstX+1]= remapX[dstX];
 rgbmap16 = malloc((1<<(SIGNIF_BITS*3))*sizeof(unsigned short));
 rgbmapBits = calloc(((1<<(SIGNIF_BITS*3))+7)/8, sizeof(unsigned char));

 /* initialise the first line */
 srcPtr = trueData;
 dstPtr = currLine;
 for (x=0; x<dstX+2; x++)
  {
   *dstPtr++ = srcPtr[remapX[x]];
   *dstPtr++ = srcPtr[remapX[x]+1];
   *dstPtr++ = srcPtr[remapX[x]+2];
  }

 for (y=0; y<dstY; y++)
  {
   /* initialise the next line */
   if (y + 1 < dstY)
    {
     /* feed in the next line */
     srcPtr = &trueData[trueBPL*(((y+1)*srcY)/dstY)];
     dstPtr = nextLine;
     if (srcX==dstX)
      {
       memcpy(&dstPtr[3], srcPtr, srcX*3);
       memcpy(&dstPtr[0], srcPtr, 3); /* duplicate first pixel */
       memcpy(&dstPtr[srcX*3+3], &srcPtr[(srcX-1)*3], 3); /* duplicate last pixel */
      }
     for (x=0; x<dstX+2; x++)
      {
       *dstPtr++ = srcPtr[remapX[x]];
       *dstPtr++ = srcPtr[remapX[x]+1];
       *dstPtr++ = srcPtr[remapX[x]+2];
      }
    }
   /* else we already have the next line equal to the last line, so no need to copy it again */

   dstPtr16 = &cmapData[cmapBPL*y];
   srcPtr = &currLine[3];
   srcPtrRight = srcPtr + 3;
   srcPtrDown = &nextLine[3];
   srcPtrDownRight = srcPtrDown + 3;
   srcPtrDownLeft =  srcPtrDown - 3;
   randPos = rand();
   for (x=dstX; x; x--)
    {
     index = ((((srcPtr[0]>>NORMALIZE_BITS)<<SIGNIF_BITS) | (srcPtr[1]>>NORMALIZE_BITS))<<SIGNIF_BITS) | (srcPtr[2]>>NORMALIZE_BITS);
     if (!(rgbmapBits[index>>3]&(1<<(index&7))))
      {
       rgbmap16[index] = (unsigned short) findClosestPaletteRGB(srcPtr, cmap, cmapColors);
       rgbmapBits[index>>3] |= (1<<(index&7));
      }
     *dstPtr16 = rgbmap16[index];
     for (c=0; c<3; ++c)
      {
       error = ((int)*srcPtr) - cmap[*dstPtr16*3+c];
       if (ditherNoise)
        error += randTable[(randPos++)&RAND_AND];
       if (ditherIntensity==16)
        {
	 *srcPtrRight     = truncateA[*srcPtrRight     + ((error*7) >> 4)];
	 *srcPtrDownLeft  = truncateA[*srcPtrDownLeft  + ((error*3) >> 4)];
	 *srcPtrDown      = truncateA[*srcPtrDown      + ((error*5) >> 4)];
	 *srcPtrDownRight = truncateA[*srcPtrDownRight + ((error*1) >> 4)];
	}
       else
        {
	 *srcPtrRight     = truncateA[*srcPtrRight     + ((error*7*ditherIntensity) >> 8)];
	 *srcPtrDownLeft  = truncateA[*srcPtrDownLeft  + ((error*3*ditherIntensity) >> 8)];
	 *srcPtrDown      = truncateA[*srcPtrDown      + ((error*5*ditherIntensity) >> 8)];
	 *srcPtrDownRight = truncateA[*srcPtrDownRight + ((error*1*ditherIntensity) >> 8)];
	}
       srcPtr++;
       srcPtrRight++;
       srcPtrDown++;
       srcPtrDownRight++;
       srcPtrDownLeft++;
      }
     dstPtr16++;
    }
   /* left right scanning does not fix the patterns... tried that, did not work... we rather use noise parameter */

   /* copy next line to current line, regardless of limits, so we duplicate the last line */
   memcpy(currLine, nextLine, (dstX+2)*3);
  }
 free(rgbmap16);
 free(rgbmapBits);
}
