/* lgif.c */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "loadimg.h"

#define MAX_OUT_CODE 5000
#define HSIZE 4096 
#define BUFFSIZE 256

typedef unsigned char BYTE;

static int BitOffset,		/* Bit Offset of next code */
  CodeSize,			/* Code size, read from GIF header */
  ReadMask,			/* Code AND mask for current code size */
  BitMask;			/* AND mask for data size */
static int *Prefix, *Suffix;	/* The hash table used by the decompressor */
static int *OutCode;		/* An output array used by the decompressor */

static FILE *fp;
static unsigned int IOResult;
static unsigned int Width;
static BYTE *Buff, *BitStream;
static unsigned int Stake;

/* skips GIF89a extensions */
static int skip89extensions(void)
{
 BYTE buf[0x100];
 int n;
 
 switch (n = fgetc(fp))
  {
   case 0x01:	/* plain text extension */
   case 0xf9:	/* graphics control extension */
   case 0xff:	/* application extension */
    n = fgetc(fp);
    if (fread(&buf[0], 1, n, fp) != n)
     return(LOAD_IO_ERROR);
    break;
   case 0xfe:	/* comment extension */
    break;	/* nothing to skip */
   default:
    return(LOAD_FILE_CORRUPT);
  }

 /* skip data blocks */
 while ((n = fgetc(fp)))
  if (fread(&buf[0], 1, n, fp) != n)
   return(LOAD_IO_ERROR);

 return(LOAD_OK);
}

/* checks if image is in selected format and reads the header */
static int tryLoad(char *filename, int *loadType, int *xpixels,
 int *ypixels, int *xdpi, int *ydpi, unsigned char *cmap, char *info,
 char *reverse, int *numInterlace, int *coreFile, int realLoad)
{
 char signature[7]; 
 unsigned int cmapSize;
 int ch;
 int byte1, byte2;

 OutCode = Prefix = Suffix = NULL; 
 Buff = BitStream = NULL;
 IOResult = LOAD_OK;

 if ((fp = fopen(filename, "rb" )) == NULL)
  return(LOAD_IO_ERROR);

 if (fread(&signature[0], 1, 6, fp) != 6)
  { fclose(fp); return(LOAD_IO_ERROR); }
 signature[6] = '\0';
 if ( (strcmp((char *) signature, "GIF87a")) &&
      (strcmp((char *) signature, "GIF89a")) )
  { fclose(fp); return(LOAD_FILETYPE_UNKNOWN); }
  
 fgetc(fp); fgetc(fp); fgetc(fp); fgetc(fp); /* skips unneeded */
 if (~(ch = fgetc(fp)) & 0x80) /* has not global color map */
  { fclose(fp); return(LOAD_WRONG_IMG); }
 cmapSize = 1 << ((ch & 0x07) + 1);
 BitMask = cmapSize - 1;
 *loadType = LOAD_CMAP; /* always return a colormap image ! */

 fgetc(fp);		/* background color - not used */
 if ((ch = fgetc(fp)))	/* aspect ratio */
   *xdpi = (*ydpi = 100000) * ((ch + 15) / 64);
 else
  *xdpi = *ydpi = 0;

 if (fread(cmap, 3, cmapSize, fp) != cmapSize)
  { fclose(fp); return(LOAD_IO_ERROR); }
  
 while ((ch = fgetc(fp)) == '!')
  if ((ch = skip89extensions()) != LOAD_OK)
   { fclose(fp); return(ch); }
 if (ch != ',')
  { fclose(fp); return(LOAD_WRONG_IMG); }
  
 fgetc(fp); fgetc(fp); fgetc(fp); fgetc(fp); /* skips unneeded data */
 byte1 = fgetc(fp);
 byte2 = fgetc(fp);
 *xpixels = Width = byte1 | (byte2 << 8);
 byte1 = fgetc(fp);
 byte2 = fgetc(fp);
 *ypixels =         byte1 | (byte2 << 8);

 if  ((*xpixels<1)||(*ypixels<1))
  { fclose(fp); return(LOAD_WRONG_IMG); }
 if ((ch = fgetc(fp)) == EOF)
  { fclose(fp); return(LOAD_IO_ERROR); }
 if (ch & 0xc0)
  { fclose(fp); return(LOAD_WRONG_IMG); }

 *coreFile = TRUE;
 *reverse = FALSE;
 *numInterlace = 1; /* no interlace */
 sprintf(info, "GIF\n%d colors", (int) cmapSize);
 return(LOAD_OK);
}

/* Fetch the next code from the raster data stream.  The codes can be
 * any length from 3 to 12 bits, packed into 8-bit bytes, so we have to
 * maintain our location in the Raster array as a BIT Offset.  We compute
 * the byte Offset into the raster array by dividing this by 8, pick up
 * three bytes, compute the bit Offset into our 24-bit chunk, shift to
 * bring the desired code to the bottom, then mask it off and return it.*/
static int ReadCode(void)
{
 unsigned int rawCode;
 unsigned int i, j, k;
 size_t bytesRead;

  k = (BitOffset >> 3) & 0x1ff;
  
  /* read new block if needed */
  if (((Stake + 0x200 - k) & 0x1ff) < 3)
   {
    if (Buff[0])	/* previous block was not last block */
     {
      i = fgetc(fp);
      if ((bytesRead = fread(Buff + 1, 1, i, fp)) != i)
       IOResult = LOAD_IO_ERROR;
      Buff[0] = (BYTE) bytesRead;
      for (i = 1, j = Stake; i <= Buff[0]; i++, j = (j + 1) & 0x1ff)
       BitStream[j] = Buff[i];
      Stake = j;
     }
    else /* attempt to read beyond last block */
     return(0);
   }

  rawCode = BitStream[k];
  if (((Stake + 0x200 - k) & 0x1ff) > 1)
   {
    rawCode |= (BitStream[(k + 1) & 0x1ff] << 8);
    if ((CodeSize >= 8) && (((Stake + 0x200 - k) & 0x1ff) > 2))
     rawCode |= BitStream[(k + 2) & 0x1ff] << 16;
   }
  rawCode >>= (BitOffset & 0x07);
  BitOffset += CodeSize;
  return (rawCode & ReadMask);
}

/* loader reads a band */
static int loadBand(unsigned char *data, int lines, int bpl)
{
 static int
   OutCount,			/* Decompressor output 'stack count' */
   InitCodeSize,		/* Starting code size, used during Clear */
   Code,			/* Value returned by ReadCode */
   MaxCode,			/* limiting value for current code size */
   ClearCode,			/* GIF clear code */
   EOFCode,			/* GIF end-of-information code */
   CurCode, OldCode, InCode,	/* Decompressor variables */
   FirstFree,			/* First free code, generated per GIF spec */
   FreeCode,			/* Decompressor, next free slot in hash table */
   FinChar;			/* Decompressor variable */
 int XC, YC;			/* Output X and Y coords of current pixel */
 static unsigned int exitJump;

 if (OutCode == NULL) /* first band ==> initialisation */
  {
   if (  ((OutCode   = (int *)  malloc((MAX_OUT_CODE+1) * sizeof(int))) == NULL)
      || ((Prefix    = (int *)  malloc(HSIZE * sizeof(int))) == NULL)
      || ((Suffix    = (int *)  malloc(HSIZE * sizeof(int))) == NULL)
      || ((Buff      = (BYTE *) malloc(BUFFSIZE)) == NULL)
      || ((BitStream = (BYTE *) malloc(0x200)) == NULL))
    return(LOAD_NO_MEMORY);

   /* Start reading the raster data. First we get the intial code size
      and compute decompressor constant values, based on this code size. */
   CodeSize = fgetc(fp);
   ClearCode = (1 << CodeSize);
   EOFCode = ClearCode + 1;
   FreeCode = FirstFree = ClearCode + 2;
   /* The GIF spec has it that the code size is the code size used to
      compute the above values is the code size given in the file,
      but the code size used in compression/decompression is the code
      size given in the file plus one. (thus the ++). */
   CodeSize++;
   InitCodeSize = CodeSize;
   MaxCode = (1 << CodeSize);
   ReadMask = MaxCode - 1;
   
   BitOffset = OutCount = 0; Stake = 0; Buff[0] = 0xff;
   Code = ReadCode();
   if (IOResult != LOAD_OK) return (IOResult);
   
   exitJump = 0;
  }

  /* Decompress the file, continuing until you see the GIF EOF code.
     One obvious enhancement is to add checking for corrupt files here. */
  XC = YC = 0; 
  while (((Code != EOFCode) || exitJump) && (YC < lines)) 
   {
    int i;
      /* Clear code sets everything back to its initial value,
         then reads the immediately subsequent code as uncompressed data. */
    if ((Code == ClearCode) && !exitJump)
     {
      CodeSize = InitCodeSize;
      MaxCode = (1 << CodeSize);
      ReadMask = MaxCode - 1;
      FreeCode = FirstFree;
      CurCode = OldCode = Code = ReadCode();
      if (IOResult != LOAD_OK)
       return (IOResult);
      FinChar = CurCode & BitMask;
      data[YC * bpl + XC] = (unsigned char) FinChar;
      if (++XC == Width) { XC = 0; YC++; }
     }
    else
     {
      /* if previous band ends in middle of the mess */
      if (exitJump) goto parashoot;
        
      /* If not a clear code, then must be data: save same as CurCode and InCode */
      CurCode = InCode = Code;
      /* If greater or equal to FreeCode, not in the hash table yet;
         repeat the last character decoded */
      if (CurCode >= FreeCode)
       {
	CurCode = OldCode;
	if (OutCount > MAX_OUT_CODE)
	 return(LOAD_FILE_CORRUPT);
	OutCode[OutCount++] = FinChar;
       }
      /* Unless this code is raw data, pursue the chain pointed to by CurCode
         through the hash table to its end; each code in the chain puts its
         associated output code on the output queue. */
      while (CurCode > BitMask)
       {
	if (OutCount > MAX_OUT_CODE)
	 return(LOAD_FILE_CORRUPT);
	OutCode[OutCount++] = Suffix[CurCode];
	CurCode = Prefix[CurCode];
       }
	/* The last code in the chain is treated as raw data. */
      FinChar = CurCode & BitMask;
      if (OutCount > MAX_OUT_CODE)
       return(LOAD_FILE_CORRUPT);
      OutCode[OutCount++] = FinChar;
      
      parashoot: exitJump = 0;
	/* Now we put the data out to the Output routine.
           It's been stacked LIFO, so deal with it that way... */
      for (i = OutCount - 1; i >= 0; i--)
       {
        data[YC * bpl + XC] = (unsigned char) OutCode[i];
	if (++XC == Width)
	 { 
	  XC = 0; 
	  if ((++YC == lines) && (i > 0))
	   {
	    OutCount = i; exitJump = 1;
	    return(LOAD_OK);
	   }
	 }
       }
      OutCount = 0;
      /* the hash table on-the-fly. No table is stored in the file. */
      Prefix[FreeCode] = OldCode;
      Suffix[FreeCode] = FinChar;
      OldCode = InCode;
      /* Point to the next slot in the table.  If we exceed the
	 current MaxCode value, increment the code size unless
	 it's already 12.  If it is, do nothing: the next code
	 decompressed better be CLEAR */
      FreeCode++;
      if (FreeCode >= MaxCode)
       {
	if (CodeSize < 12)
	 {
	  CodeSize++;
	  MaxCode *= 2;
	  ReadMask = (1 << CodeSize) - 1;
	 }
       }
     }
    Code = ReadCode();
    if (IOResult != LOAD_OK) return (IOResult);
  }
 return(LOAD_OK);
}

/* closes the input file and releases any temporary storage */
static int endLoad()
{
 if (OutCode   != NULL) free(OutCode);
 if (Prefix    != NULL) free(Prefix);
 if (Suffix    != NULL) free(Suffix);
 if (Buff      != NULL) free(Buff);
 if (BitStream != NULL) free(BitStream);
 fclose(fp);
 return(LOAD_OK);
}

/* register GIF image loader */
void registerLoadGIF(void)
{
 registerLoader(*tryLoad, *loadBand, *endLoad);
}
