/***********************************************************************
**  Purpose:	PCX/DCX loader
**  Adapted by Simon & Goran & Dusan
**  Version:	1.0
**  Date:	April 1991
**  Author:	James D. Murray, Anaheim, CA, USA
**  C Compilers:	Borland C++ v2.0, Microsoft C v6.00a
**  Copyright (C) 1991 by Graphics Software Labs. All rights reserved.
************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "loadimg.h"

#define PCXL

#define PCX_ID      0
#define PCX_VER     1
#define PCX_ENC     2
#define PCX_BPP     3
#define PCX_XMINL   4
#define PCX_XMINH   5
#define PCX_YMINL   6
#define PCX_YMINH   7
#define PCX_XMAXL   8
#define PCX_XMAXH   9
#define PCX_YMAXL   10
#define PCX_YMAXH   11
/* hres (12,13) and vres (14,15) not used */
#define PCX_CMAP    16 /* start of 16*3 colormap data */
#define PCX_PLANES  65
#define PCX_BPRL    66
#define PCX_BPRH    67

#define PCX_MAPSTART 0x0c /* Start of appended colormap */

/* PCX Run-length Decoder
** PCX decoders need to be aware of image data that contains runs which
** start at the end of one scan line and finish at the start of the next
** line.  This function decodes one scan line at a time and will keep
** track of runs if they extend to the next scan line.
** Decode (uncompress) a PCX image file scan line.
** PcxDecodeScanLine() decodes a buffer containing scan line data encoded
** using the PCX run-length encoding algorithm.  The encoded data is read
** from a FILE stream, decoded, and then written to a pointer to a buffer
** passed to this function.
** The PCX specification states (in so many words) that the run-length
** encoding of a pixel run should stop at the end of each scan line.
** However, some PCX encoders may continue the encoding of a pixel run on
** to the beginning of the next scan line, if possible.  This code,
** therefore, assumes that any pixel run can span scan lines.
** To check for decoding errors, the value returned by PcxDecodeScanLine()
** should be the same as the value of BufferSize (the length of an
** uncompressed scan line).
** Returns: Total number of pixels decoded from compressed scan line,
**          or EOF if a file stream error occured. */

static unsigned char runcount;	/* Length of decoded pixel run */
static unsigned char runvalue;	/* Value of decoded pixel run */

static int PcxDecodeScanLine (unsigned char *DecodedBuffer, int BufferSize, FILE * FpPcx)
/* DecodedBuffer;    Pointer to buffer to hold decoded data */
/* int  BufferSize;  Size of buffer to hold decoded data */
/* FILE *FpPcx;      FILE pointer to the open input PCX image file */
{
  int index = 0;	/* Index into compressed scan line buffer */
  int total = 0;	/* Running total of decoded pixel values */
  unsigned char byte;	/* Data byte read from PCX file */

  /* If there is any data left over from the previous scan
     line write it to the beginning of this scan line. */
  do
    { /* Write the pixel run to the buffer */
      for (total += runcount;	/* Update total */
	   runcount && index < BufferSize; /* Don't read past buffer */
	   runcount--, index++)
	DecodedBuffer[index] = runvalue; /* Assign value to buffer */

      if (runcount)		/* Encoded run ran past end of scan line */
	{
	  total -= runcount;	/* Subtract count not written to buffer */
	  return (total);	/* Return number of pixels decoded */
	}

      /* Get the next encoded run packet.
      ** Read a byte of data.  If the two MBSs are 1 then this byte
      ** holds a count value (0 to 63) and the following byte is the
      ** data value to be repeated.  If the two MSBs are 0 then the
      ** count is one and the byte is the data value itself.*/
      byte = (unsigned char) getc(FpPcx); /* Get next byte */

      if ((byte & 0xC0) == 0xC0) /* Two-byte code */
	{
	  runcount = (unsigned char) (byte & 0x3F); /* Get run count */
	  runvalue = (unsigned char) getc(FpPcx); /* Get pixel value */
	}
      else /* One byte code */
	{
	  runcount = 1;	/* Run count is one */
	  runvalue = byte; /* Pixel value */
	}
    }
  while (index < BufferSize); /* Read until the end of the buffer */

  if (ferror(FpPcx))
    return (EOF); /* File stream error.  Probably a premature EOF */

  return (total); /* Return number of pixels decoded */
}

/* This code is an example of several functions that work with PCX image
** files.  PCXCODE reads a PCX image file, decodes the image, data
** re-encodes it, and writes it out to a second PCX file.  While this
** program is only a working example, it can be used to possibly repair
** PCX files that have been previously encoded incorrectly and fail to
** display properly. */

static FILE *fp;		/* PCX image file input FILE stream  */
static unsigned char *DecodedBuffer;/* Buffer holding decoded scan line */
static int NumOfColors;		/* Number of colors in the image */
static int ScanLineLength;	/* Length of uncompressed scan line  */
static int Width;		/* Width of image in pixels */
static unsigned char pcxPlanes=1;

/* checks if image is in selected format and reads the header */
static int tryLoad(char *fname, int *loadType, int *xpixels,
 int *ypixels, int *xdpi, int *ydpi, unsigned char *cmap, char *info,
 char *reverse, int *numInterlace, int *coreFile, int realLoad)
{
  register int i;		/* Loop counter */
  int xSize;			/* Width of image in pixels */
  int ySize;			/* Height of image in scan lines */
  unsigned char PcxVgaPal[768];	/* PCX VGA palette structure */
  int IOResult;
  unsigned char pcxHead[128]; /* PCX image file header structure */

  runcount = runvalue = 0;
  
  /* Open the input PCX image file */
  if ((fp = fopen (fname, "rb")) == (FILE *) NULL)
   return(LOAD_IO_ERROR);

  /* Read the PCX image file header information */
  IOResult = fread (pcxHead, 128, 1, fp);

  /* Check for FILE stream error */
  if (ferror (fp) || (IOResult != 1))
   { fclose(fp); return (LOAD_IO_ERROR); }

  /* Check the identification byte value */
  if ((pcxHead[PCX_ID] != 0x0A) || (pcxHead[PCX_VER] > 5))
   { fclose(fp); return (LOAD_FILETYPE_UNKNOWN); }
  pcxPlanes = pcxHead[PCX_PLANES];

  /* Calculate size of image in pixels and scan lines */
  xSize = (pcxHead[PCX_XMAXL] + ((int) pcxHead[PCX_XMAXH] << 8))
    - (pcxHead[PCX_XMINL] + ((int) pcxHead[PCX_XMINH] << 8));

  ySize = (pcxHead[PCX_YMAXL] + ((int) pcxHead[PCX_YMAXH] << 8))
    - (pcxHead[PCX_YMINL] + ((int) pcxHead[PCX_YMINH] << 8));

  xSize++;
  ySize++;

  *xpixels = Width = xSize;
  *ypixels = ySize;
  if ((*xpixels>MAX_LOAD_IMAGE) || (*ypixels>MAX_LOAD_IMAGE) || (*xpixels<1) || (*ypixels<1))
   { fclose(fp); return (LOAD_FILETYPE_UNKNOWN); }

  NumOfColors = 1 << (pcxHead[PCX_BPP] * pcxHead[PCX_PLANES]);
  ScanLineLength = (pcxHead[PCX_BPRL] + ((int) pcxHead[PCX_BPRH] << 8));
  ScanLineLength *= pcxHead[PCX_PLANES];

//printf("PCX num of colors %d version %d\n", NumOfColors, pcxHead[PCX_VER]);
  /* If there is a VGA palette then read and append it to the output file */
  if (pcxHead[PCX_VER] == 5)
   {
    long int savedPos = ftell(fp);
    
    fseek(fp, -769L, SEEK_END); /* Seek backwards from end of file */
    if (getc (fp) == 0x0C)	/* VGA Palette ID value? */
     {
      /* Read the VGA palette */
      for (i = 0; i < sizeof (PcxVgaPal); i++)
       PcxVgaPal[i] = (unsigned char) getc(fp);
      memcpy(cmap, &PcxVgaPal[0], 768);
     }
    else
     if (NumOfColors<=256)
      memcpy(cmap, &pcxHead[PCX_CMAP], NumOfColors * 3);
    fseek(fp, savedPos, SEEK_SET);
   }
  else
   if (NumOfColors<=256)
    memcpy(cmap, &pcxHead[PCX_CMAP], NumOfColors * 3);

  if (NumOfColors<=256)
   *loadType = LOAD_CMAP; /* always prepare a cmap image, even if b&w */
  else
   *loadType = LOAD_TRUE_COL;
  *xdpi = *ydpi = 0;	/* no information on dpi */
  DecodedBuffer = NULL;
  sprintf(info, "PCX\n%d colors", NumOfColors);
  *reverse = FALSE;
  *numInterlace = 1; /* no interlace */
  *coreFile = TRUE;
//printf("file %s %dx%d\n", fname, *xpixels, *ypixels);
  return (LOAD_OK);
}	

/* loader reads a band */
static int loadBand(unsigned char *data, int lines, int bpl)
{
 register int i,j;
 unsigned char *dataP;

 if (DecodedBuffer == NULL)	/* first band */
  if ((DecodedBuffer = (unsigned char *) malloc(ScanLineLength)) == NULL)
   return (LOAD_NO_MEMORY);
   
 if (NumOfColors == 2)
  for (i = 0; i < lines; i++)
   {
    dataP = (unsigned char *) data + bpl * i;     
    if (PcxDecodeScanLine(DecodedBuffer, ScanLineLength, fp)
	!= ScanLineLength)
     return (LOAD_FILE_CORRUPT);
    else
     for (j = 0; j < Width; j++)
       *dataP++ = (DecodedBuffer[j>>3] >> (7 - (j&7))) & 1;
   }
 else
  for (i = 0; i < lines; i++)
   {
    dataP = (unsigned char *) data + bpl * i;
    if (PcxDecodeScanLine(DecodedBuffer, ScanLineLength, fp)
	!= ScanLineLength)
     return (LOAD_FILE_CORRUPT);
    else
     {
      int scan_div_planes = ScanLineLength / (int) pcxPlanes;
      unsigned char plane;

      if (NumOfColors == 256)
       memcpy(dataP, DecodedBuffer, Width);
      else
       for (j = 0; j < Width; j++)
	{
	 *dataP = 0;
	 for (plane = 0; plane < pcxPlanes; plane++)
	  *dataP |= (unsigned char) ((
	   (DecodedBuffer[(j>>3) + scan_div_planes * plane] >> (7 - (j&7)))
	   & 1) << plane);
	 dataP++;
	}
     }
   }
 return(LOAD_OK);
}	

/* closes the input file and releases any temporary storage */
static int endLoad()
{
 if (DecodedBuffer != NULL)
  free ((char *) DecodedBuffer);
 fclose (fp);
 return(LOAD_OK);
}

/* register PCX image loader */
void registerLoadPCX()
{
 registerLoader(*tryLoad, *loadBand, *endLoad);
}
