/***************************************************************************
                          apaintImageResize.cpp  -  description
                             -------------------
    begin                : Tue Jun 29 2004
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
// ==========================================================
// Upsampling / downsampling classes
//
// Design and implementation by
// Herve Drolon (drolon@infonie.fr)
// Source has been taken from FreeImage Project and changed for use in ArahPaint.
// by Matjaz Gutenberger

#include "apaintImageResize.h"
#include <stdlib.h>
#include <qcolor.h>

/**
  Filter weights table.
  This class stores contribution information for an entire line (row or column).
*/
CWeightsTable::CWeightsTable(CGenericFilter *pFilter, int uDstSize, int uSrcSize, double dScale)
{
 int u;
 double dWidth;
 double dFScale = 1.0;
 double dFilterWidth = pFilter->GetWidth();

  if(dScale < 1.0) {
   // minification
   dWidth = dFilterWidth / dScale;
   dFScale = dScale;
  } else {
   // magnification
   dWidth = dFilterWidth;
  }

  // allocate a new line contributions structure
  //
  // window size is the number of sampled pixels
  m_WindowSize = 2 * (int)ceil(dWidth) + 1;
  m_LineLength = uDstSize;
  // allocate list of contributions
  m_WeightTable = (Contribution*)malloc(m_LineLength * sizeof(Contribution));
  for(u = 0 ; u < m_LineLength ; u++)
  {
   // allocate contributions for every pixel
    m_WeightTable[u].Weights = (double*)malloc(m_WindowSize * sizeof(double));
  }


  for(u = 0; u < m_LineLength; u++)
  {
     // scan through line of contributions
     double dCenter = (double)u / dScale;   // reverse mapping
     // find the significant edge points that affect the pixel
     int iLeft = QMAX (0, (int)floor (dCenter - dWidth));
     int iRight = QMIN ((int)ceil (dCenter + dWidth), int(uSrcSize) - 1);

     // cut edge points to fit in filter window in case of spill-off
     if((iRight - iLeft + 1) > int(m_WindowSize))
     {
       if(iLeft < (int(uSrcSize) - 1 / 2)) {
         iLeft++;
       } else {
         iRight--;
       }
     }

     m_WeightTable[u].Left = iLeft;
     m_WeightTable[u].Right = iRight;

     int iSrc = 0;
     double dTotalWeight = 0.0;  // zero sum of weights
     for(iSrc = iLeft; iSrc <= iRight; iSrc++)
     {
       // calculate weights
       double weight = dFScale * pFilter->Filter(dFScale * (dCenter - (double)iSrc));
       m_WeightTable[u].Weights[iSrc-iLeft] = weight;
       dTotalWeight += weight;
     }
     if((dTotalWeight > 0) && (dTotalWeight != 1))
     {
       // normalize weight of neighbouring points
       for(iSrc = iLeft; iSrc <= iRight; iSrc++) {
         // normalize point
        m_WeightTable[u].Weights[iSrc-iLeft] /= dTotalWeight;
       }
     }
   }
}

CWeightsTable::~CWeightsTable()
{
  for(int u = 0; u < m_LineLength; u++)
  {
    // free contributions for every pixel
    free(m_WeightTable[u].Weights);
  }
  // free list of pixels contributions
  free(m_WeightTable);
}

// ---------------------------------------------

/**
 CResizeEngine<br>
 This class performs filtered zoom. It scales an image to the desired dimensions with
 any of the CGenericFilter derived filter class.<br>
 It only works with 8-bit buffers.<br><br>

 <b>References</b> : <br>
 [1] Paul Heckbert, C code to zoom raster images up or down, with nice filtering.
 UC Berkeley, August 1989. [online] http://www-2.cs.cmu.edu/afs/cs.cmu.edu/Web/People/ph/heckbert.html
 [2] Eran Yariv, Two Pass Scaling using Filters. The Code Project, December 1999.
 [online] http://www.codeproject.com/bitmap/2_pass_scaling.asp

*/

ApaintImage CResizeEngine::resize(ApaintImage &src, int dstWid, int dstHei, ResizeFilterType ftyp)
{
  int x, y, bpp;
  int channel;
  uchar *src_bits, *dst_bits;
  ApaintImage src8, dst8, dst;
  QRgb *ctab =0;


  if (src.type() == LOAD_CMAP)
   ctab = src.colorTable();
  bpp = (src.depth()/8);
  if ((dstWid <= 0) || (dstHei <= 0)) 
    return src.copy(); //return null image
  if ((src.width() == dstWid) && (src.height() == dstHei))
    return src.copy();

  // select filter
  CGenericFilter *pFilter = NULL;
  switch(ftyp)
  {
    case FILTER_SMOOTH2:
    case FILTER_SMOOTH3:
    case FILTER_SMOOTH4:
    case FILTER_NONE:
     return src.scale(dstWid, dstHei);
    case FILTER_BOX:
      pFilter = new CBoxFilter();
    break;
   case FILTER_BICUBIC:
      pFilter = new CBSplineFilter; //CBicubicFilter();
    break;
   case FILTER_BILINEAR:
      pFilter = new CBilinearFilter();
    break;
   case FILTER_BSPLINE:
      pFilter = new CBSplineFilter();
    break;
   case FILTER_CATMULLROM:
      pFilter = new CCatmullRomFilter();
    break;
   case FILTER_LANCZOS3:
      pFilter = new CLanczos3Filter();
    break;
  }

  CResizeEngine Engine(pFilter);

  // allocate a temporary 8-bit dib (no need to build a palette)
  int width  = src.width();
  int height = src.height();

  src8.newImage(width, height, 8);
  dst.newImage(dstWid, dstHei, 24);
  src.copyImageDataTo(&dst);
  // process each channel separately
  // -------------------------------   

  for(channel = 0; channel < 3; channel++)
   {
    // extract channel from source dib
    for(y = 0; y < height; y++)
     {
      src_bits = src.scanLine(y);
      dst_bits = src8.scanLine(y);
      if (src.type() == LOAD_CMAP)
       {
        for(x = 0; x < width; x++)
         {
          switch (channel)
           {
            case 0:
              dst_bits[x] = qRed(ctab[src_bits[channel]]);
              break; 
            case 1:
              dst_bits[x] = qGreen(ctab[src_bits[channel]]);
              break;
            case 2:
              dst_bits[x] = qBlue(ctab[src_bits[channel]]);
              break;
           }
          src_bits += bpp;
         }
      }
     else
      if (src.type() == LOAD_TRUE_COL)
       {
        for(x = 0; x < width; x++)
         {
          dst_bits[x] = src_bits[channel];
          src_bits += bpp;
         }
       }
     }

    // process channel
    dst8 = Engine.scale(src8, dstWid, dstHei);

    // insert channel to destination dib
    for(y = 0; y < dstHei; y++)
     {
      src_bits = dst8.scanLine(y);
      dst_bits = dst.scanLine(y);
      for(x = 0; x < dstWid; x++)
       {
        dst_bits[channel] = src_bits[x];
        dst_bits += 3;
       }
     }
   }

  delete pFilter;
  return dst;
  
}

ApaintImage CResizeEngine::scale(ApaintImage &src, int dst_width, int dst_height)
{
  int src_width  = src.width();
  int src_height = src.height();

  // allocate the dst 8-bit image
  ApaintImage dst;
  dst.newImage(dst_width, dst_height, 8);
  if(dst.isNull())
    return dst;

  // decide which filtering order (xy or yx) is faster for this mapping by
  // counting convolution multiplies

  if(dst_width*src_height <= dst_height*src_width)
  {
    // xy filtering
    // -------------

    // allocate a temporary image
    ApaintImage tmp;
    tmp.newImage(dst_width, src_height, 8);
    if(tmp.isNull())
      return tmp;


     // scale source image horizontally into temporary image
     horizontalFilter(src, src_width, src_height, tmp, dst_width, src_height);

     // scale temporary image vertically into result image
     verticalFilter(tmp, dst_width, src_height, dst, dst_width, dst_height);


  }
  else
  {
    // yx filtering
    // -------------

    // allocate a temporary image
    ApaintImage tmp;
    tmp.newImage(src_width, dst_height, 8);
    if(tmp.isNull())
      return tmp;

    // scale source image vertically into temporary image
    verticalFilter(src, src_width, src_height, tmp, src_width, dst_height);

    // scale temporary image horizontally into result image
    horizontalFilter(tmp, src_width, dst_height, dst, dst_width, dst_height);
   }
	return dst;
}


/// Performs horizontal image filtering
void CResizeEngine::horizontalFilter(ApaintImage &src, int src_width, int src_height, ApaintImage &dst, int dst_width, int dst_height)
{
  if(dst_width == src_width)
  {
    // no scaling required, just copy 
    dst = src.copy();      
  }
  else
  {
   // allocate and calculate the contributions
   CWeightsTable weightsTable(m_pFilter, dst_width, src_width, double(dst_width) / double(src_width));

   // step through rows
   for(int y = 0; y < dst_height; y++)
   {
     // scale each row
     uchar *src_bits = src.scanLine(y);
     uchar *dst_bits = dst.scanLine(y);

     for(int x = 0; x < dst_width; x++)
     {
       // loop through row
       double value = 0;
       int iLeft = weightsTable.getLeftBoundary(x);    // retrieve left boundary
       int iRight = weightsTable.getRightBoundary(x);  // retrieve right boundary

       for(int i = iLeft; i <= iRight; i++)
       {
         // scan between boundaries
         // accumulate weighted effect of each neighboring pixel
         double weight = weightsTable.getWeight(x, i-iLeft);
         value += (weight * (double)src_bits[i]);
       }

       // place result in destination pixel
       dst_bits[x] = (uchar)QMIN(QMAX((int)0, (int)(value + 0.5)), (int)255);
     }
    }
  }
}

/// Performs vertical image filtering
void CResizeEngine::verticalFilter(ApaintImage &src, int src_width, int src_height, ApaintImage &dst, int dst_width, int dst_height)
{
  if(src_height == dst_height)
  {
    // no scaling required, just copy
    for(int y = 0; y < src_height; y++)
    {
      dst = src.copy();
    }
  }
  else
  {
   // allocate and calculate the contributions
   CWeightsTable weightsTable(m_pFilter, dst_height, src_height, double(dst_height) / double(src_height));

   // step through columns
   for(int x = 0; x < dst_width; x++)
   {
     // scale each column
     for(int y = 0; y < dst_height; y++)
     {
       // loop through column
       double value = 0;
       int iLeft = weightsTable.getLeftBoundary(y);    // retrieve left boundary
       int iRight = weightsTable.getRightBoundary(y);  // retrieve right boundary

       uchar *dst_bits = dst.scanLine(y);

       for(int i = iLeft; i <= iRight; i++)
       {
         // scan between boundaries
         // accumulate weighted effect of each neighboring pixel
         uchar *src_bits = src.scanLine(i);
         double weight = weightsTable.getWeight(y, i-iLeft);
         value += (weight * (double)src_bits[x]);
       }

       // clamp and place result in destination pixel
       dst_bits[x] = (uchar)QMIN(QMAX((int)0, (int)(value + 0.5)), (int)255);
     }
    }
  }
}

