/***************************************************************************
                          toolBezier.cpp  -  description
                             -------------------
    begin                : Wed Feb 26 2003
    copyright            : (C) 2003-2011 Arahne, d.o.o.
    email                : arahne@arahne.si
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cmath>
#include "apaintconfig.h"
#include "toolBezier.h"
#include "apaintdrawrepeat.h"
#include "apaintUtil.h"
#include "apaint.h"
#include "apaintdoc.h"

using namespace std;

char *getCommonPNG(const char *pixName);

BezierTool::BezierTool(DrawInRepeat *drawp, ApaintDoc*doc, BezierToolProp *p, PropObject *prop):ToolBase(DrawTool, prop)
{
 this->p = p;
 cur=QCursor(ArrowCursor);
 setToggleBtn(QPixmap(getCommonPNG("draw-bezier")), tr("Bezier curve"));
 cur=QCursor(ArrowCursor);
 setToolChangeType(NoChange);
 this->doc = doc;
 Q_CHECK_PTR(drawp);
 draw = drawp;
 resetTool();
 isActiveTool = false;
 oldLineWidth = 0;
}

BezierTool::~BezierTool()
{
}

void BezierTool::showCommonProp(bool enable)
{
 CommonProp *cProp = CommonProp::getObject();
 cProp->setAspectDrawProp(enable);
 cProp->setThickLinesProp(enable);

 if (enable) p->show();
 else p->hide();
}

void BezierTool::begin()
{
 isActiveTool = true;
 resetTool();
}

void BezierTool::end()
{
 isActiveTool = false;
 if (points.size() > 2)
 {
  if (state == Enter) points.resize(points.size()-3);
  state = Finish;
  emit repaintSig();
 }

 doc->refreshLayer();
 clearStacks();
 ToolBase::end();
}

void BezierTool::afterRefresh(QPainter *pc)
{
 if (state != Finish) drawPoints(pc, points);
 else resetTool();
}

void BezierTool::resetTool()
{
 selectedPoint = -1;
 selectedNode = -1;
 points.resize(0);
 oldPoints.resize(0);
 state = Enter;

// if (doc && doc->getMainLayer())
// {
//	 float w = doc->getMainLayer()->imageBuf.width();
//	 float h = doc->getMainLayer()->imageBuf.height();
//
//	 if (w>0 && h>0)
//	 {
//		 int pts = 500*3+2;
//		 points.resize(pts);
//
//		 for (int i=0; i<pts; i++)
//		 {
//			 int x = rand()*w/RAND_MAX;
//			 int y = rand()*h/RAND_MAX;
//			 points[i] = QPoint(x,y);
//		 }
//	 }
//	 lineWidth = 1;
//	 oldLineWidth = 0;
// }

 while (undoStack.isEmpty() == false) delete undoStack.pop();

 actionComplete();
 emit clearStacks();
 p->goEnterMode();
}

void BezierTool::drawPoints(QPainter *pc, const QPointArray &points)
{
 int size = oldPoints.size();
 QPointArray pts(size);
 int first = (selectedNode+3)/3*3-2;
 int second = (selectedNode+2)/3*3-1;

 pc->setRasterOp(XorROP);
 pc->setPen(QColor(0,128,255));

 for (int i=0; i<size; ++i) pts[i] = QPoint((oldPoints[i].x()+0.5)*zoomX, (oldPoints[i].y()+0.5)*zoomY);

 if (selectedNode != -1 && points.size() > 2)
  {
   if (first > 0 && first < size-2)
   {
	pc->drawLine(pts[first-1], pts[first]);
    pc->setPen(QColor(128,0,255));

	for (int i=-3; i<=3; i++)
	 pc->drawLine(pts[first]+QPoint(i,-3), pts[first]+QPoint(i,3));

    pc->setPen(QColor(0,128,255));
   }
   if (second > 1 && second < size-2)
   {
	pc->drawLine(pts[second], pts[second+1]);
    pc->setPen(QColor(128,0,255));

	for (int i=-3; i<=3; i++)
	 pc->drawLine(pts[second]+QPoint(i,-3), pts[second]+QPoint(i,3));
   }
  }

 pc->setBrush(QBrush(QColor(255,128,0)));

 for (int pos = 0; pos < size; pos++)
  if (pos % 3 == 0 || pts.size() == 2)
   {
    pc->setPen(QColor(255,128,0));
	pc->drawLine(pts[pos]+QPoint(-3,-1), pts[pos]+QPoint(-3,1));
	pc->drawLine(pts[pos]+QPoint(-2,-2), pts[pos]+QPoint(-2,2));
	pc->drawLine(pts[pos]+QPoint(-1,-3), pts[pos]+QPoint(-1,3));
	pc->drawLine(pts[pos]+QPoint(0,-3), pts[pos]+QPoint(0,3));
	pc->drawLine(pts[pos]+QPoint(1,-3), pts[pos]+QPoint(1,3));
	pc->drawLine(pts[pos]+QPoint(2,-2), pts[pos]+QPoint(2,2));
	pc->drawLine(pts[pos]+QPoint(3,-1), pts[pos]+QPoint(3,1));
   }
}

void BezierTool::actionComplete()
{
 UndoRedoObject *uobj = new UndoRedoObject;
 undoStack.push(uobj);

 actionStarted = false;
 uobj->lineWidth = lineWidth;
 uobj->state = state;
 uobj->selectedNode = selectedNode;
 uobj->selectedPoint = selectedPoint;
 uobj->points.resize(points.size());

 for (uint i=0; i<points.size(); i++)
	 uobj->points[i] = points[i];

 while (redoStack.isEmpty() == false) delete redoStack.pop();

 emit putToStack();
}

#define MAXIMAL_DISTANCE 10

void BezierTool::calcPointClick(QPoint mousePosition)
{
 if (points.size() == 0) return;

 int size = controlPoints();
 int tmp[2] = { (selectedNode+3)/3*3-2, (selectedNode+2)/3*3-1 };
 float min = MAXIMAL_DISTANCE*MAXIMAL_DISTANCE*1.0;

 if (selectedNode == -1) tmp[0] = tmp[1] = -1;

 selectedPoint = -1;

 if (size > 2)
 {
  for (int i=0; i<2; i++)
   if (tmp[i] >= 0 && tmp[i] < size)
   {
    float dx = mousePosition.x() - points[tmp[i]].x();
    float dy = mousePosition.y() - points[tmp[i]].y();
    float d = dx*dx + dy*dy;

    if (d < min)
    {
     min = d;
     selectedPoint = tmp[i];
    }
   }

  if (selectedPoint == -1)
   for (int i=0; i<size; i+=3)
   {
    float dx = mousePosition.x() - points[i].x();
    float dy = mousePosition.y() - points[i].y();
    float d = dx*dx + dy*dy;

    if (d < min)
    {
     min = d;
     selectedNode = i;
     selectedPoint = i;
    }
   }
 }
 else if (size == 2)
 {
  for (int i=0; i<2; i++)
  {
   float dx = mousePosition.x() - points[i].x();
   float dy = mousePosition.y() - points[i].y();
   float d = dx*dx + dy*dy;

   if (d < min)
   {
    min = d;
    selectedPoint = i;
   }
  }
 }

 if (selectedPoint != -1)
 {
  if (selectedNode % 3 == 0 && selectedNode == selectedPoint && points.size() > 2)
  {
   QPoint vec = mousePosition - points[selectedNode];

   if (selectedNode > 0) points[selectedNode-1] += vec;
   if (selectedNode+3 < points.size()) points[selectedNode+1] += vec;
   points[selectedNode] = mousePosition;
  }
  else if (selectedPoint < points.size())
  {
   points[selectedPoint] = mousePosition;
  }
 }
}

QPoint BezierTool::calcPoint(int i, float t0)
{
 if (points.size() == 2) return points[0] + (points[1] - points[0])*t0;
 if (points.size() < 2 || i < 0 || i+3 >= points.size()) return QPoint();

 float t1 = 1 - t0;
 float ax = points[i].x();
 float ay = points[i].y();
 float bx = points[i+1].x();
 float by = points[i+1].y();
 float cx = points[i+2].x();
 float cy = points[i+2].y();
 float dx = points[i+3].x();
 float dy = points[i+3].y();
 float ax1 = ax*t1+bx*t0, bx1 = bx*t1+cx*t0, cx1 = cx*t1+dx*t0;
 float ax2 = ax1*t1+bx1*t0, bx2 = bx1*t1+cx1*t0;
 float x = ax2*t1+bx2*t0;
 float ay1 = ay*t1+by*t0, by1 = by*t1+cy*t0, cy1 = cy*t1+dy*t0;
 float ay2 = ay1*t1+by1*t0, by2 = by1*t1+cy1*t0;
 float y = ay2*t1+by2*t0;

 return QPoint(round(x), round(y));
}

void BezierTool::moveSegment(QPoint mousePosition, float t)
{
 if (points.size() < 2) return;

 if (points.size() == 2)
 {
  QPoint p = calcPoint(0, t);
  QPoint vec = mousePosition - p;

  points[0] += vec;
  points[1] += vec;
 }
 else
 {
  int i = selectedNode/3*3;

  if (i<0 || i+3>=points.size()) return;

  float t1 = 1 - t;
  QPoint p = (mousePosition - t1*t1*t1*points[i] - t*t*t*points[i+3])/(3*t*t1);
  QPoint b = p - t*movingVector;
  QPoint c = p + t1*movingVector;

  points[i+1] = b;
  points[i+2] = c;
 }
}

float BezierTool::lineSegmentDistance(QPoint mousePosition, float &t)
{
 if (points.size() != 2) return MAXIMAL_DISTANCE;

 t = -1;
 int x, y;

 if (points[0] != points[1])
 {
  float bax = points[1].x() - points[0].x();
  float bay = points[1].y() - points[0].y();
  float max = mousePosition.x() - points[0].x();
  float may = mousePosition.y() - points[0].y();

  t = (max*bax+may*bay)/(bax*bax+bay*bay);

  if (t<0 || t>1) return MAXIMAL_DISTANCE;

  x = points[0].x() + bax*t;
  y = points[0].y() + bay*t;
 }
 else
 {
  x = points[0].x();
  y = points[0].y();
  t = 0.5;
 }

 float dx = mousePosition.x() - x;
 float dy = mousePosition.y() - y;
 float d = sqrt(dx*dx+dy*dy);

 return d;
}

float BezierTool::curveDistance(QPoint mousePosition, float &t)
{
 if (points.size() <= 2) return MAXIMAL_DISTANCE;

 float t0;
 float min = MAXIMAL_DISTANCE*MAXIMAL_DISTANCE;
 float mx = mousePosition.x();
 float my = mousePosition.y();

 t = -1;

 for (uint i = 0; i+3<points.size(); i+=3)
 {
  for (t0=1.0/64; t0<1; t0+=1.0/64)
  {
   QPoint p = calcPoint(i, t0);

   float d2 = (p.x()-mx)*(p.x()-mx)+(p.y()-my)*(p.y()-my);

   if (d2 < min)
   {
	selectedNode = i+1;
	selectedPoint = i+1;
	movingVector = points[selectedPoint+1]-points[selectedPoint];
	t = t0;
	min = d2;
   }
  }
 }

 return sqrt(min);
}

float BezierTool::distance(QPoint mousePosition, float &t)
{
 if (points.size() == 2) return lineSegmentDistance(mousePosition, t);
 else return curveDistance(mousePosition, t);
}

void BezierTool::insertPoint(QPoint mousePosition, float t)
{
 if (points.size() < 2)
 {
  points.resize(2);
  points[0] = mousePosition;
  points[1] = mousePosition;
 }
 else if (points.size() == 2)
 {
  points.putPoints(2, 3, 0, 0, points[1].x(), points[1].y(), 0, 0);
  points[1] = (mousePosition+points[0])*0.5;
  points[2] = (mousePosition+points[3])*0.5;
  points[4] = (3*points[3]-mousePosition)*0.5;
  selectedPoint = -1;
  selectedNode = -1;
 }
 else
 {
  selectedNode = selectedNode/3*3;
  float t0 = t;
  float t1 = (1-t);
  float f0 = t0;
  float f1 = t1;
  int i = selectedNode;

  QPoint a1 = points[i];
  QPoint d2 = points[i+3];
  QPoint b1 = a1 - (points[i]-points[i+1])*f0;
  QPoint c2 = d2 - (points[i+3]-points[i+2])*f1;

  f0 *= t0;
  f1 *= t1;

  QPoint c1 = -a1 + 2*b1 + (points[i] - 2*points[i+1] + points[i+2])*f0;
  QPoint b2 = -d2 + 2*c2 + (points[i+3] - 2*points[i+2] + points[i+1])*f1;

  f0 *= t0;

  QPoint d1 = a1 - 3*b1 + 3*c1 - (points[i] - 3*points[i+1] + 3*points[i+2] - points[i+3])*f0;

  points.putPoints(points.size(), 3, 0, 0, 0, 0, 0, 0);
  selectedNode += 3;

  for (int i=points.size()-1; i>selectedNode+2; --i) points[i] = points[i-3];

  points[selectedNode-2] = b1;
  points[selectedNode-1] = c1;
  points[selectedNode] = d1;
  points[selectedNode+1] = b2;
  points[selectedNode+2] = c2;
  selectedPoint = selectedNode;
 }
}

void BezierTool::removePoint()
{
 if (points.size()<=2 || selectedNode==-1) return;

 for (uint i=selectedNode-(selectedNode > 0); i<points.size()-3; i++) points[i] = points[i+3];

 points.resize(points.size()-3);
 selectedPoint = -1;
}

namespace
{
 float t = -1;
}

void BezierTool::mousePressed(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
 QPoint mousePosition = QPoint(e->x()/zoomX, e->y()/zoomY);

 if (e->stateAfter() == Qt::LeftButton)
 {
  t = -1;

  if (state == Enter)
  {
   if (points.size() == 0)
   {
	   insertPoint(mousePosition, 0);
       actionComplete();
   }
   else actionStarted = true;
//   uobj.what = INSERT_NODE;
  }
  else
  {
   calcPointClick(mousePosition);

   if (selectedPoint != -1) actionStarted = true;
   else
   {
	if (points.size() == 2)
	{
	 if (lineSegmentDistance(mousePosition, t) < MAXIMAL_DISTANCE)
	 {
		moveSegment(mousePosition, t);
		actionStarted = true;
	 }
	}
	else if (points.size() > 2)
	{
	 if (curveDistance(mousePosition, t) < MAXIMAL_DISTANCE)
	 {
		moveSegment(mousePosition, t);
		actionStarted = true;
	 }
	}
   }
  }
 }
 else if (e->stateAfter() == Qt::MidButton)
 {
  float t;

  calcPointClick(mousePosition);

  if (selectedPoint != -1 && selectedPoint == selectedNode)
  {
   removePoint();
   actionComplete();
  }
  else if (distance(mousePosition, t) < MAXIMAL_DISTANCE)
  {
   insertPoint(mousePosition, t);
   actionComplete();
  }
 }
 else if (e->stateAfter() == RightButton)
 {
  if (state == Enter && points.size() >= 2)
  {
   if (points.size() > 2)
   {
	points.resize(points.size()-3);
   }
   state = Edit;
   p->goEditMode();
  }
  else if (state == Edit)
  {
	state = Finish;
    p->goEnterMode();
  }
 }

 repaint(pc, pi);
}

void BezierTool::mouseMoved(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
 QPoint mousePosition(e->x()/zoomX, e->y()/zoomY);

 if (e->stateAfter() == Qt::LeftButton)
  {
   if (state == Enter)
    {
     if (points.size() > 2)
      {
       int pos = points.size() - 2;

       points[pos] = mousePosition;
       points[pos+1] = 2*points[pos] - points[pos-1];
      }
     else if (points.size() == 2) points[1] = mousePosition;
    }
   else if (state == Edit)
    {
	 if (points.size() == 2 && t == -1 && selectedPoint != -1)
	  points[selectedPoint] = mousePosition;
	 else if (t > 0 && t < 1) moveSegment(mousePosition, t);
	 else if (selectedNode != -1)
	 {
	  if (selectedPoint == selectedNode && selectedNode % 3 == 0)
	  {
	   QPoint vec = mousePosition - points[selectedPoint];

	   if (selectedPoint > 0) points[selectedPoint-1] += vec;
	   if (selectedPoint < points.size()-1) points[selectedPoint+1] += vec;
	   points[selectedPoint] += vec;
	  }
	  else if (selectedPoint >= 0) points[selectedPoint] = mousePosition;
	 }
    }
  }
 else if (e->stateAfter() == 0 && points.size() > 2)
 {
  if (state == Enter)
  {
   points[points.size()-1] = points[points.size()-2] = points[points.size()-3] = mousePosition;
  }
 }

 repaint(pc, pi);
}

void BezierTool::mouseReleased(QMouseEvent *e, QPainter *pc, QPainter *pi)
{
 int x = e->pos().x() / zoomX;
 int y = e->pos().y() / zoomY;

 if (state == Enter && points.size() >= 2)
 {
  points.putPoints(points.size(), 3, x, y, x, y, x, y);
 }

 if (actionStarted) actionComplete();

 repaint(pc, pi);
}

void BezierTool::repaint(QPainter *pc, QPainter *pi)
{
 static QValueList<int> drawingSegments;
 static QValueList<int> erasingSegments;

 erasingSegments.clear();
 drawingSegments.clear();

 if (points.size() < 2) return;
 if (state != Finish && oldPoints == points && oldLineWidth == lineWidth) return;

 if (state == Finish || lineWidth != oldLineWidth || (points.size()+oldPoints.size() < 10))
 {
  if (oldPoints.size() == 2) erasingSegments.append(0);
  if (points.size() == 2) drawingSegments.append(0);
  for (uint i=0; i+2<oldPoints.size(); i += 3) erasingSegments.append(i);
  for (uint i=0; i+2<points.size(); i += 3) drawingSegments.append(i);
 }
 else
 {
  if (points.size() < oldPoints.size())
  {
   if (selectedNode == -1) erasingSegments.append(oldPoints.size()-5);
   else
   {
    if (selectedNode >= 3) erasingSegments.append(selectedNode-3);
    if (selectedNode < oldPoints.size()) erasingSegments.append(selectedNode);
    if (selectedNode >= 3 && selectedNode < points.size()) drawingSegments.append(selectedNode-3);

    selectedNode = selectedPoint = -1;
   }
  }
  else if (points.size() > oldPoints.size())
  {
   if (state == Enter) drawingSegments.append(points.size()-5);
   else
   {
    erasingSegments.append(selectedNode-3);
    drawingSegments.append(selectedNode-3);
    drawingSegments.append(selectedNode);
   }
  }
  else
  {
   if (state == Enter)
   {
	erasingSegments.append(oldPoints.size()-5);
	drawingSegments.append(points.size()-5);
   }
   else if (selectedPoint == selectedNode && selectedNode % 3 == 0)
   {
	if (selectedNode >= 3)
	{
	 erasingSegments.append(selectedNode-3);
	 drawingSegments.append(selectedNode-3);
	}

	if (selectedNode+3 < points.size())
	{
	 erasingSegments.append(selectedNode);
	 drawingSegments.append(selectedNode);
	}
   }
   else if (selectedNode % 3 == 0)
   {
	if (selectedPoint < selectedNode)
	{
     erasingSegments.append(selectedNode-3);
     drawingSegments.append(selectedNode-3);
	}
	else
	{
     erasingSegments.append(selectedNode);
     drawingSegments.append(selectedNode);
	}
   }
   else
   {
	int i = selectedNode/3*3;

	erasingSegments.append(i);
	drawingSegments.append(i);
   }
  }
 }

 draw->drawBezier(pi, oldPoints, points, erasingSegments, drawingSegments, oldLineWidth, lineWidth, state == Finish);

// QPen pen = pi->pen();
//
// pen.setWidth(oldLineWidth);
// pi->setPen(pen);
// pi->setRasterOp(Qt::CopyROP);
// draw->drawBezier(pi, oldPoints, erasingSegments, false, false, false);
// pi->setRasterOp(CopyROP);
// pen.setWidth(lineWidth);
// pi->setPen(pen);
// draw->drawBezier(pi, points, drawingSegments, true, true, state == Finish);
 oldLineWidth = lineWidth;

 if (state == Finish) resetTool();
 else
 {
  uint size = points.size();

  oldPoints.resize(size);

  for (uint i=0; i<size; i++) oldPoints[i] = points[i];

  doc->refreshLayer();
 }
}

int BezierTool::controlPoints() const
{
 int size = points.size();

 return size <= 2 ? size : --size;
}

bool BezierTool::goEnterMode()
{
 if (state != Edit) return false;

 state = Finish;

 return true;
}

bool BezierTool::goEditMode()
{
 if (state != Enter || points.isEmpty()) return false;

 state = Edit;

 if (points.size() > 2) points.resize(points.size()-3);

 return true;
}

void BezierTool::slotUndo()
{
 if (isActiveTool == false || undoStack.isEmpty()) return;

 UndoRedoObject *uobj = undoStack.pop();

 if (undoStack.isEmpty())
 {
  undoStack.push(uobj);

  return;
 }

 if (state != undoStack.top()->state)
 {
	 if (undoStack.top()->state == Edit) p->goEditMode();
	 else p->goEnterMode();
 }

 redoStack.push(uobj);
 lineWidth = undoStack.top()->lineWidth;
 selectedNode = undoStack.top()->selectedNode;
 selectedPoint = undoStack.top()->selectedPoint;
 state = (State) undoStack.top()->state;
 points.resize(undoStack.top()->points.size());

 for (uint i=0; i<undoStack.top()->points.size(); i++)
	 points[i] = undoStack.top()->points[i];

 repaintSig();
}

void BezierTool::slotRedo()
{
 if (isActiveTool == false || redoStack.isEmpty()) return;

 UndoRedoObject *uobj = redoStack.top();
 undoStack.push(uobj);

 if (state != uobj->state)
 {
	 if (uobj->state == Edit) p->goEditMode();
	 else p->goEnterMode();
 }

 lineWidth = uobj->lineWidth;
 selectedNode = uobj->selectedNode;
 selectedPoint = uobj->selectedPoint;
 state = (State) uobj->state;
 points.resize(uobj->points.size());

 for (uint i=0; i<uobj->points.size(); i++) points[i] = uobj->points[i];

 redoStack.pop();
 repaintSig();
}

void BezierTool::newDocSlot()
{
 resetTool();
}

void BezierTool::lineSizeChgSlot(const QPen &pen)
{
 lineWidth = pen.width();

 if (isActiveTool)
 {
	 actionComplete();
	 emit repaintSig();
 }
}

BezierToolProp::BezierToolProp(QWidget *parent, const char *name) : QButtonGroup(parent)
{
 setCaption("");
 setColumnLayout(0, Qt::Vertical);
 QVBoxLayout *vbox = new QVBoxLayout(layout());
 layout()->setSpacing(1);
 layout()->setMargin(1);
 vbox->setAlignment(Qt::AlignTop);
 enterMode = new QRadioButton(this);
 editMode = new QRadioButton(this);
 vbox->addWidget(enterMode);
 vbox->addWidget(editMode);
 enterMode->setText(ApaintConfig::getText(794));
 editMode->setText(ApaintConfig::getText(25));
 enterMode->toggle();
 enterMode->show();
 editMode->show();
 connect(enterMode, SIGNAL(pressed()), this, SLOT(enterModePressed()));
 connect(editMode, SIGNAL(pressed()), this, SLOT(editModePressed()));
}

void BezierToolProp::goEnterMode()
{
 if (enterMode->isChecked() == false)
 {
  editMode->toggle();
  enterMode->toggle();
 }
}

void BezierToolProp::goEditMode()
{
 if (editMode->isChecked() == false)
 {
  enterMode->toggle();
  editMode->toggle();
 }
}

void BezierToolProp::enterModePressed()
{
 if (enterMode->isChecked()) enterMode->toggle();
 else
 {
  if (bezierTool->goEnterMode()) editMode->toggle();
  else enterMode->toggle();
 }
}

void BezierToolProp::editModePressed()
{
 if (editMode->isChecked()) editMode->toggle();
 else
 {
  if (bezierTool->goEditMode()) enterMode->toggle();
  else editMode->toggle();
 }
}
